# mwprop.ne2001p v1.0 Jan 2024

"""
Constants needed for interstellar wave propagation and other applications.
All units are cgs unless indicated otherwise

Note: could keep the constants as 'fully-fledged Quantity objects'
as stated at https://docs.astropy.org/en/stable/constants/
But then each usage would have to extract a value. 
Maybe that would be ok at the beginning of modules that 
need specific constants.

Example: c = constants.c  instead of c = constants.c.cgs.value
and then: c = constants.c.cgs.value if c needs to be cgs in a module

If done this way, would put evaluations of DM and RM constants, etc. 
into a separate module (perhaps)

Address this later.
"""
from __future__ import print_function
# 30 Jun 2021 added Plank's constant
# 31 Dec 2019

from math import pi
from math import sqrt
from astropy import constants

# fundamental

eV = 1.60218e-12    # erg
c = constants.c.cgs.value
m_e = constants.m_e.cgs.value
e_e = constants.e.esu.value
r_e = e_e**2 / (m_e * c**2)

m_p = constants.m_p.cgs.value

h = constants.h.cgs.value 
hbar = constants.hbar.cgs.value 

kb = constants.k_B.cgs.value

# lengths

cm_in_m = 100.
km = 1.e5        # cm
au = constants.au.cgs.value
pc = constants.pc.cgs.value
kpc = 1.e3*pc
Mpc = 1.e6*pc
Gpc = 1.e9*pc

# angular
mas = pi / (180.*3600.*1000.)
uas = mas / 1000.
all_sky_sr = 4.*pi
all_sky_degsq = (180./pi)**2 * all_sky_sr

# frequencies

GHz = 1.e9
MHz = 1.e6

lambda1 = c / GHz        # wavelength in cm for 1 GHz

# temporal
ns = 1.e-9       # sec
mus = 1.e-6      # sec
ms = 1.e-3       # sec

# Dispersion constant
KDM = (c * r_e * pc) / (2.*pi*GHz**2)                # for TOAs in seconds

# Faraday rotation measure constant

microGauss = 1.e-6

# RMconstant gives standard rad m^{-2} 
# for B in microgauss, pathlength in pc, electron density in cm^{-3}

RMconstant = ((e_e**3) / (2.*pi*m_e**2 * c**4)) * pc * cm_in_m**2 * microGauss
RMconstant_cgs = ((e_e**3) / (2.*pi*m_e**2 * c**4)) * pc * microGauss

# Scattering measure kpc m^{-20/3} to cgs 
SMunit = kpc*10.**(-40./3.)
SM1 = 1.        # kpc m^{-20/3}
SM_cgs = SM1 * kpc * cm_in_m**(-20./3.)     # same as SMunit (a check)

# Flux density

Jy = 1.e-23     # cgs
Jansky = Jy

# Brightness temperature for 1 Jy pulse of 1 ms at 1 kpc at 1 GHz
Tb111 = (Jy * kpc**2) / (2.*kb*(GHz*ms)**2)

# plasma frequency in Hz for unit electron density
ne1 = 1.        # 1/cc
nup = sqrt((4.*pi * ne1 * e_e**2) / m_e) / (2.*pi)      # Hz

# cyclotron frequency in a 1 Gauss field
B1=1.
nuc = (e_e * B1) / (m_e * c) / (2.*pi)                  # Hz




# mwprop.nemod v2.0 Jan 2026

"""
ne_arms
Replacement for Fortran routine NE_ARMS_LOG_MOD in density.NE2001.f

Version history:
2020 Jan 19
01/19/20 --- JMC
     * initial conversion f77 --> python
02/08/20 --- JMC
     * removed rsun = 8.5
     * all model parameters now imported from config_ne2001p
"""

from mwprop.nemod.config_nemod import *

script_path = os.path.dirname(os.path.realpath(__file__))

def ne_arms_ne2001p(x,y,z, Ncoarse=20, dthfine=0.01, nfinespline=5, verbose=False):
    """
    Evaluates electron density and F parameter for arm nearest to x,y,z

    Input:
        x, y, z = location in NE2001  coordinates  in kpc
        Ncoarse = number of coarse samples along each arm 
        dthfine = step size in Galactocentric angle for fine sampling (rad)
        nfinespline = number of coarse samples to use for fine-sampling spline
        verbose = True:  writes out n_e component information 

    Output:
        ne        electron density        cm^{-3}
        F         F parameter             composite units
        whicharm  which spiral arms       1 to 5, no close arm ==> 0 
    """

    # First find coarse location of points on arms nearest to input point

    narms = np.shape(coarse_arms)[1]
    dsq_coarse = (coarse_arms[0] - x)**2 + (coarse_arms[1] - y)**2
    index_dsqmin = [dsq_coarse[j, :].argmin() for j in range(narms)]

    # Zoom in to find precision location and evaluate density and F parameter

    # number of coarse samples to use as input to fine spline
    nspline2 = int((nfinespline-1)/2)     
    
    # Initialize
    nea = 0.
    ga = 0.
    whicharm = 0
    whicharm_spiralmodel = 0

    # Cylindrical radius and Galactocentric angle of input point;
    # thxydeg measured CCW from +y axis

    rr = sqrt(x**2 + y**2)
    thxydeg = np.rad2deg(np.arctan2(-x, y))

    if verbose:
        print('arms rr, thxydeg: ', rr, thxydeg)

    for j in range(narms):

        # Use fine spline on coarse samples  to find nearest position on j-th arm

        ind1 = range(max(0, index_dsqmin[j]-nspline2-1), 
                     min(index_dsqmin[j]+nspline2+1, Ncoarse), 1)
        dsqs = CubicSpline(th1[j, ind1], dsq_coarse[j, ind1])    
        thjfine = np.arange(th1[j, ind1[0]], th1[j, ind1[-1]], dthfine)
        ind_min = dsqs(thjfine).argmin()
        thjmin = thjfine[ind_min]
        sj = CubicSpline(th1[j,:], r1[j,:])                     
        rjmin = sj(thjmin)
        xjmin, yjmin = -rjmin*np.sin(thjmin), rjmin*np.cos(thjmin)

        # Evaluate electron density for nearest spiral arm if it is within
        # some multiple of the e-folding half-width of the arm
   
        dmin = sqrt((x-xjmin)**2 + (y-yjmin)**2)
        jj = Darmmap[str(j)]
        wa = Dgal['wa'] 

        """
        Note armmap used here is to maintain the legacy arm numbering
        from NE2001 and TC93 using as input the arm numbering in Wainscoat.
        New NE model could dispense with this. 
        """

        if thxydeg < 0: thxydeg += 360
        if j== 0: dmin_min = dmin
        Wa = wa * Dgal['warm'+str(jj)]
        if dmin < 3.*wa:
        #if dmin < 5.*wa:               # This helps reduce discontinuities
            if dmin <= dmin_min:
                dmin_min = dmin
                whicharm_spiralmodel = j+1

            # arm-distance factor:
            argxy = dmin / Wa
            ga = np.exp(-argxy**2)

            # Galactocentric radial factor:
            if rr > Dgal['Aa']: ga *= sech2((rr-Dgal['Aa'])/2.)  

            # z factor:
            Ha = Dgal['ha'] * Dgal['harm'+str(jj)]
            ga *= sech2(z/Ha)

            # Amplitude re-scalings as in NE2001 code;
            # Ignore cases where these have been turned off in that code (fossils!).
            
            # TC arm 3:
            if jj == 3:
                th3adeg, th3bdeg = 290, 363 
                test3 = thxydeg - th3adeg
                if test3 < 0: test3 += 360
                if 0 <= test3 and test3 < th3bdeg-th3adeg:
                    arg = 2.*pi*test3 / (th3bdeg-th3adeg)
                    fac = ((1 + np.cos(arg))/2)**4
                    # TEMP: fac = 1
                    ga *= fac

            # TC arm 2:
            if jj == 2:
                th2adeg, th2bdeg = 340, 370
                test2 = thxydeg - th2adeg
                fac2min = 0.1
                if test2 < 0: test2 += 360
                if 0 <= test2 and test2 < th2bdeg-th2adeg:
                    arg = 2.*pi*test2 / (th2bdeg-th2adeg)
                    fac = ((1 + fac2min + (1 - fac2min) * np.cos(arg))/2)**3.5
                    # fac = 1    # TEMP
                    ga *= fac
            
            nea += ga * Dgal['narm'+str(jj)] * Dgal['na'] 
            s = sqrt(x**2 + (rsun-y)**2)
            #s = sqrt(x**2 + (8.5-y)**2)
            if verbose:
                print('arms: ', j, jj, whicharm_spiralmodel, index_dsqmin[j], 
                    max(0, index_dsqmin[j]-nspline2-1),
                    min(index_dsqmin[j]+nspline2+1, Ncoarse), ind1)
                print(
                    '%5.2f  %5.2f  %d  %d  %d    %4.2f  %4.2f  %4.2f    %4.2f  %7.5f'
                     %(s, rr, j, jj, whicharm_spiralmodel, wa, Wa, dmin, ga, nea))
    if whicharm_spiralmodel == 0:
        whicharm = 0
        Farm = 0
    else:
        whicharm = Darmmap[str(whicharm_spiralmodel-1)]
        #Farm = Dgal['Fa'] * Dgal['farm'+str(whicharm)]    # Intended value

        ###### NOTE ######
        # 2020 Feb 9:  
        # found that the Fortran code doesn't use Farm calculated using 
        # Fa x farm_j parameters.  This seems to be an error in dmdsm.NE2001.f, 
        # which uses Fa to calculate SM for the spiral arms instead of Faval,
        # the value returned by density.NE2001.f.    This only affects the
        # spiral arm values for SM.   

        # For now, maintain this error in the Python code because the aim here
        # is to replicate the Fortran code:  

        Farm = Dgal['Fa']
 
        # The question is then whether the error was in the code when fitting 
        # was done to find the best values of farm_j?   I think the error was 
        # not there during the fitting because some of the earlier code versions
        # use Fa * farm_j.    

    if verbose:
        print('arms whicharm_spiralmodel  whicharm: ', 
            whicharm_spiralmodel, whicharm)
    
    return nea, Farm, whicharm


# mwprop v2.0 Jan 2026

"""
dmdsm.py
NE20x integrator

Line of sight integration routines for NE20x. 

Separate routines for calculating distance D from input dispersion measure DM
and vice verse.

This version is for both NE2025 and NE2001p = Python version of NE2001 (fortran).
JMC 2021 Dec 28 - 2022 Jan 05
SKO 2022 Mar - 2023 Nov
JMC 2023 Dec 31 

This version includes:
    1. Use of coarse sampling + spline interpolation of smooth density components 
       (ne1, ne2, nea from thin and thick disks and spiral arm components)
    2. Fine sampling on all other components (LISM, GC, clumps, and voids)
    3. Ancillary analyses of the line of sight are done only if do_analysis = True
       in calling argument of dmdsm_dm2d in order to minimize computation times. .

Based on dmdsm_dm2d_ne2001_dev.py

Note mwprop needs to be in PYTHONPATH
"""
#import os
#script_path = os.path.dirname(os.path.realpath(__file__))

import numpy as np

if int(np.__version__[0]) >=2:
    from numpy import trapezoid as trapz
else:
    from numpy import trapz

from numpy import log10
from numpy import array, linspace, where, size
from numpy import digitize, interp
from scipy.interpolate import interp1d

from mwprop.nemod.config_nemod import *
from mwprop.nemod.density import *
from mwprop.nemod.neclumpN_fast import *

import datetime as datetime
import time

import sys,os

import warnings
original_showwarning = warnings.showwarning
def _warning(message,category = UserWarning,filename = '',lineno = -1,file=None,line=None):
    print('Warning: ',message)
warnings.showwarning = _warning

script_path = os.path.dirname(os.path.realpath(__file__))
basename = sys.argv[0].split('/')[-1].split('.')[0]
now = datetime.datetime.now()
plotstamp = basename + '_' + str(now).split('.')[0]

# set which model
#inpdir = os.path.dirname(os.path.realpath(__file__))+'/params/which_model.inp'
#which_mod = np.loadtxt(inpdir,dtype=str)
#if which_mod=='NE2001':
#	eval_NE2001 = True
#	eval_NE2025 = False
#if which_mod=='NE2025':
#	eval_NE2025 = True
#	eval_NE2001 = False

#print(eval_NE2025,eval_NE2001)

# ----------------------------------------------------------------------

def calc_galcentric_vecs(l, b, dmax, Ns):
    """
    Calculates vectors for line of sight and galactocentric coordinates
    Input:
        l,b = Galactic longitude and latitude (rad)
        dmax = maximum distance along LoS
        Ns = number of steps along LoS
    Output:
        svec = vector of steps along the LoS up to dmax
        xvec, yvec, zvec = vectors of x,y,z coordinates
    """
    sl=sin(l)
    cl=cos(l)
    sb=sin(b)
    cb=cos(b)

    svec = linspace(0, dmax, Ns)
    rvec = svec * cb
    xvec = rvec * sl
    yvec = rsun - rvec * cl
    zvec = svec * sb
    return svec, xvec, yvec, zvec

# ----------------------------------------------------------------------
# SKO -- 3/2/22 changed ds_coarse to 0.1 for both functions (was previously 0.2 for dm2d), changed ds_fine to 0.005
# changed Nsmin to 20 from 10
# SKO -- 3/6/22 -- changed ds_fine back to 0.01, issue was actually in nevoidN_NE2001.py

def dmdsm_dm2d(l, b, dm_target, ds_coarse=0.1, ds_fine=0.01, Nsmin=20, 
    dm2d_only = False, do_analysis=True, plotting=False, verbose=False, debug=True):
    """
    Integrates electron density from NE20x model to reach the target DM.
    in the direction expressed in Galactic coordinates. 

    Computes pulsar distance and scattering measure
    from model of Galactic electron distribution.

    Input: l         galactic longitude in radians
           b         galactic latitude in radians
           dm_target input value DM (dispersion measure in pc/cm^3)
           ds_coarse coarse step size along LoS (kpc)
           ds_fine   fine step size along LoS (kpc)
           Nsmin     minimum number of samples to use along LoS
           dm2d_only True => calculate only the distance; otherwise also calculate SM.
           do_analysis True => analyze components of line of sight: dm, sm, lism, arms
           verbose   prints a few diagnostics

    Output:
           limit  (set to '>' if only a lower distance limit can be
               given; otherwise set to ' ')
           dist          calculated distance or input distance
           dmpsr         calculated DM or input DM
           sm            scattering measure, uniform weighting) (kpc/m^{20/3} 
           smtau         scattering measure, weighting for pulse broadening 
           smtheta       scattering measure, weighting for angular broadening
                          of galactic sources 
           smiso          scattering measure appropriate for calculating the
                           isoplanatic angle at the source's location
           Uses Kolmogorov spectral index = 11/3
           Useful constants: c_sm = (sikol - 3) / (2 * (2*pi)**(4-sikol)) 
                             sm_factor = c_sm * units_conversion to kpc m^{-20/3} 
                             (defined in config_ne2001p.py)       
    """
    sm_iso_index = sikol - 2                # should be 5/3 for Kolmogorov index sikol=11/3
    limit=' '                       

    # Do multiple passes on integration:
    # Pass  0: coarse integration to get dmax to use for fine steps
    # Pass >0: distance estimate using fine distance steps

    dm_reached = False

    npasses = 3
    mult_dmaxnom = 1.5          # multiplier of dhat to guard against undershooting

    npass = 0
    dhat = 0
    dm_calc_max = 0
    while  npass < npasses and dm_reached is False:
 
        if npass==0:
            # Calculate nominal maximum distance to calculate initial set of samples
            # Use small nominal n_e for small DM, larger for large DM
            # Rationale:  large DM requires inner Galaxy LoS that sample larger n_e
            ne_nom = 0.01 * (1 + log10(dm_target))
            dmax_integrate = \
                 min(mult_dmaxnom * dm_target/ ne_nom / pc_in_kpc, dmax_ne2001p_integrate)
            Ns_coarse = int(dmax_integrate/ds_coarse)
            if Ns_coarse < Nsmin:
                Ns_coarse = Nsmin
                ds_coarse = dmax_integrate/ Ns_coarse
            Ns_fine = int(dmax_integrate/ds_fine)
            if Ns_fine < Nsmin:
                Ns_fine = Nsmin
                ds_fine = dmax_integrate / Ns_fine
                
            # Identify clumps that contribute to this LoS:
            # SKO 11/23 -- rcmult now defined in config_ne2001.py; rcmult smaller for clumps with edge = 0 (hard cutoff) 
            relevant_clump_indices = relevant_clumps(l, b, dmax_integrate, rcmult)
           
            if debug:
                print(np.size(relevant_clump_indices), 
                    ' relevant clumps out of ', nclumps, ' total')
                print(relevant_clump_indices)
        else:
            dmax_integrate = min(mult_dmaxnom * dhat, dmax_ne2001p_integrate)
            Ns_fine = int(dmax_integrate / ds_fine) 
            if Ns_fine < Nsmin:
                Ns_fine = Nsmin 
                ds_fine = dmax_integrate / Ns_fine
        if debug:
            print('dmax_integrate = ', dmax_integrate)
            print('Ns_fine, ds_fine = ', Ns_fine, ds_fine)
            print('ds_coarse = ', ds_coarse, ' ds_fine = ', ds_fine)
            print('Ns_coarse = ', Ns_coarse, ' Ns_fine = ', Ns_fine)
 
        sc_vec, xc_vec, yc_vec, zc_vec = calc_galcentric_vecs(l, b, dmax_integrate, Ns_coarse)
        sf_vec, xf_vec, yf_vec, zf_vec = calc_galcentric_vecs(l, b, dmax_integrate, Ns_fine)

        # Obtain electron density components, F parameters, weights, etc.
        # Use separate calls to density_2001_smooth_comps and density_2001_smallscale_comps

        # ----------------------------------------------
        # Smooth, large-scale components on coarse grid: 
        # ----------------------------------------------
        # Note only cnd_smooth, cFsmooth needed here
        cne1,cne2,cnea, cF1, cF2, cFa, cwhicharm, cne_smooth, cFsmooth = \
            array([
               density_2001_smooth_comps(xc_vec[j],yc_vec[j],zc_vec[j]) 
               for j in range(Ns_coarse)
            ]).T

        # Spline functions:
        cs_ne_smooth = CubicSpline(sc_vec, cne_smooth)
        cs_F_smooth = CubicSpline(sc_vec, cFsmooth)

        # Evaluate smooth components on fine grid  using spline functions:
        ne_smooth = cs_ne_smooth(sf_vec)
        F_smooth = cs_F_smooth(sf_vec)

        # Resample cwhicharm using digitize: use coarse vec as bins for fine vec:
        inds_whicharm = np.digitize(sf_vec, sc_vec, right=True)    
        whicharm = cwhicharm[inds_whicharm] 

        # ------------------------------------
        # Small-scale components on fine grid:
        # ------------------------------------
        negc,nelism,necN,nevN, Fgc, Flism, FcN, FvN, wlism, wldr, wlhb, wlsb, wloopI, \
           hitclump, hitvoid, wvoid = \
               array([
                   density_2001_smallscale_comps(\
                       xf_vec[j],yf_vec[j],zf_vec[j], inds_relevant=relevant_clump_indices) \
                       for j in range(Ns_fine)\
               ]).T

        #if debug:
        #	print('hitvoid',hitvoid)
        wtotal = (1-wgvN*wvoid)*(1-wglism*wlism)        # used for SM calculations
        ne_ex_clumps_voids = (1.-wglism*wlism) * (ne_smooth + wggc*negc) + wglism*wlism*nelism
        ne = (1-wgvN*wvoid)*ne_ex_clumps_voids  + wgvN*wvoid*nevN + wgcN*necN 
        #if np.count_nonzero(wgvN*wvoid*nevN)!=0:
        	#print('Warning: Void(s) intersected. Run los_diagnostics.py for details.')
        #if np.count_nonzero(wgcN*necN)!=0:
        	#print('Warning: Clump(s) intersected. Run los_diagnostics.py for details.')

        dm_cumulate_vec = \
            pc_in_kpc * array([trapz(ne[:j], sf_vec[:j]) for j in range(1, Ns_fine+1) ])
        dm_calc_max = dm_cumulate_vec[-1]       # maximum dm calculated for this pass

        # Interpolate to get distance estimate:
        dhat = interp(dm_target, dm_cumulate_vec, sf_vec)

        if debug: 
            print('Pass ', npass, '  Ns_fine = %d  dmax_integrate = %5.2f   dhat = %5.2f'\
                %(Ns_fine, dmax_integrate, dhat))
        if dm_calc_max < dm_target: # SKO -- modified warning
            #print('Warning: Pass %d    terminal DM < target DM (%6.1f, %6.1f)   dhat = %6.2f' %(npass, dm_calc_max, dm_target, dhat)) 
            if npass+1 == npasses:
                warnings.warn('terminal DM < target DM')
            dhat *= mult_dmaxnom                # increase dhat to reach target DM
        else:
            dm_reached = True
        npass += 1

    # Convert flags to integers:
    whicharm = whicharm.astype(int)
    hitclump = hitclump.astype(int)
    hitvoid = hitvoid.astype(int)

    if np.count_nonzero(hitclump)!=0:
    	warnings.warn('Clump(s) intersected. Run los_diagnostics.py for details.')
    if np.count_nonzero(hitvoid)!=0:
    	warnings.warn('Void(s) intersected. Run los_diagnostics.py for details.')

    if debug:
    	print('ne',ne)
    	#print('hitvoid:',hitvoid)
    # Integrate using trapz to get cumulative DM:
    dm_cumulate_vec = pc_in_kpc * array([trapz(ne[:j], sf_vec[:j]) for j in range(1,Ns_fine+1)])
    dm_calc_max = dm_cumulate_vec[-1]       # maximum dm calculated 

    if debug:
    	print('dm_calc_max',dm_calc_max)
    # Test if calculated DMs reach dm_target.
    # If not, attribute this to a lower bound on the distance, as in NE2001 (fortran),
    # though it might be better to attribute this to insufficent electrons.
    # In new version NE2001x, use this second approach. 

    if dm_calc_max < dm_target:     # did not reach target DM; find dhat lower limit
     
        limit = '>'
        indlim = where(dm_cumulate_vec==dm_cumulate_vec[-1])[0][0]
        dm_reached = dm_cumulate_vec[indlim]
        dhat_lower_limit = sf_vec[indlim]
        dhat = dhat_lower_limit     # for consistency with return statements
        dm_return = dm_reached
        dhat_return = dhat_lower_limit
       
        #if debug:
            #indmin = where(dm_cumulate_vec==dm_cumulate_vec[-1])[0][0]
            #print('indmin = ', indmin, sf_vec[indmin])
    else:                           # Reached target DM; interpolate to get distance estimate:
        dhat = interp(dm_target, dm_cumulate_vec, sf_vec)
        dhat_return = dhat
        dm_return = dm_target


    if plotting:
        plot_dm_along_LoS(
            dm_target, dhat, sf_vec, dm_cumulate_vec, relevant_clump_indices, dc, 
            plot_dm_target=True, saveplot=True)

    if dm2d_only is True:
        if do_analysis:

            if eval_NE2001:
                savedir = os.getcwd()+'/output_ne2001p/'
            elif eval_NE2025:
                savedir = os.getcwd()+'/output_ne2025p/'
            f24outfile  = savedir+'f24_dm2d_only_lism_etc_paths' + '.txt'
            f24 = open(f24outfile, 'w')

            f25outfile  = savedir+'f25_dm2d_only_ne_vs_s' + '.txt'
            f25 = open(f25outfile, 'w')

            sf_vec, ne, ne1, ne2, nea = \
                analysis_dmd_dm_only(f24, f25,
                    l, b, dm_target, dhat, sf_vec, xf_vec, yf_vec, zf_vec, sc_vec,
                    cne1, cne2, cnea, ne, whicharm, hitclump, hitvoid, dm_cumulate_vec,
                    nelism, negc, necN, nevN, wtotal, wlism, wvoid,
                    wlhb, wldr, wlsb, wloopI)
        if debug:
            return limit, dhat_return, dm_return, sf_vec, dm_cumulate_vec
        else:
            return limit, dhat_return, dm_return

    else:           # then also calculate scattering measure, do_analysis, etc.

        # Calculate dSM (proportional to Cn2) from individual dSM components
        # Note: dsm1, dsm2, dsma are subsumed by dsm_smooth
        # dsm1, dsm2, dsma are calculated in the `analysis' code block (if requested)

        # For now, the dsm quantities are not multiplied by sm_factor.
        # That is done later in calculating sm quantities from dsm quantities.
        # Could change this to make it a bit more transparent
        # e.g. multiply dsm quantities by sm_factor here.

        dsm_smooth = wtotal * ne_smooth**2 * F_smooth
        dsmgc = wtotal*wggc*negc**2*Fgc
        dsmlism = (1.-wgvN*wvoid)*wglism*wlism*nelism**2*Flism
        dsmcN = wgcN*necN**2*FcN
        dsmvN = wgvN*wvoid*nevN**2*FvN
        dsm = dsm_smooth + dsmgc + dsmlism + dsmcN + dsmvN

        # Calculate integrals needed to evaluate SM.
        # Integrate from s = 0 to s = dhat (starting from observer's position at s = 0). 
        # First integrate over sf_vec then use cubic spline to find SM at d = dhat.

        Nsf1 = np.size(sf_vec) + 1
        dsm_cumulate1_vec = array([trapz(dsm[:j], sf_vec[:j]) for j in range(1, Nsf1)])
        dsm_cumulate2_vec = \
            array([trapz(sf_vec[:j]*dsm[:j], sf_vec[:j]) for j in range(1, Nsf1)])
        dsm_cumulate3_vec = \
            array([trapz(sf_vec[:j]**2*dsm[:j], sf_vec[:j]) for j in range(1, Nsf1)])
        dsm_cumulate4_vec = \
            array([trapz(sf_vec[:j]**sm_iso_index*dsm[:j], sf_vec[:j]) for j in range(1, Nsf1)])

        sm_cumulate = sm_factor * dsm_cumulate1_vec
        smtau_cumulate = 6 * (sm_factor/dhat) * (dsm_cumulate2_vec - dsm_cumulate3_vec/dhat)
        smtheta_cumulate = 3 * sm_factor * \
            (dsm_cumulate1_vec + dsm_cumulate3_vec/dhat**2 - 2*dsm_cumulate2_vec/dhat)
        smiso_cumulate = sm_factor * dsm_cumulate4_vec

        # cubic splines:

        sm_cs = CubicSpline(sf_vec, sm_cumulate)
        smtau_cs = CubicSpline(sf_vec, smtau_cumulate)
        smtheta_cs = CubicSpline(sf_vec, smtheta_cumulate)
        smiso_cs = CubicSpline(sf_vec, smiso_cumulate)

        sm_hat = sm_cs(dhat)
        smtau_hat = smtau_cs(dhat)
        smtheta_hat = smtheta_cs(dhat)
        smiso_hat = smiso_cs(dhat)

        if plotting:
            plot_dm_sm_along_LoS(
                dm_target, dm_cumulate_vec, dhat, sf_vec, 
                sm_hat, smtau_hat, smtheta_hat, smiso_hat,
                dsm, sm_cumulate, smtau_cumulate, smtheta_cumulate, smiso_cumulate,
                saveplot=True)

        if do_analysis:

            if eval_NE2001:
                savedir = os.getcwd()+'/output_ne2001p/'
            elif eval_NE2025:
                savedir = os.getcwd()+'/output_ne2025p/'
            f24outfile  = savedir+'f24_dm2d_lism_etc_paths' + '.txt'
            f24 = open(f24outfile, 'w')

            f25outfile  = savedir+'f25_dm2d_ne_dsm_vs_s' + '.txt'
            f25 = open(f25outfile, 'w')

            svec, ne, ne1, ne2, nea = \
            analysis_dmd_dm_and_sm(f24, f25,
                l, b, dm_target, dhat, sf_vec, xf_vec, yf_vec, zf_vec, sc_vec,
                cne1, cF1, cne2, cF2, cnea, cFa,
                sm_hat, ne, dsm, whicharm, hitclump, hitvoid, dm_cumulate_vec,
                nelism, negc, necN, nevN, wtotal, wlism, wvoid,
                dsmgc, dsmlism, dsmcN, dsmvN,
                wlhb, wldr, wlsb, wloopI)

        if debug:
            return limit, dhat_return, dm_return, sm_hat, smtau_hat, smtheta_hat, smiso_hat, \
               sf_vec, dm_cumulate_vec
        else:
            return limit, dhat_return, dm_return, sm_hat, smtau_hat, smtheta_hat, smiso_hat

# ----------------------------------------------------------------------

def dmdsm_d2dm(l, b, d_target, ds_coarse, ds_fine, Nsmin, 
    d2dm_only=False, do_analysis=True, plotting=False, verbose=False): # ds_coarse, ds_fine, Nsmin defined in config_ne2001

    """
    Integrates electron density from NE2001 model out to a specified
    distance  in the direction expressed in Galactic coordinates. 

    Computes pulsar distance and scattering measure
    from model of Galactic electron distribution.

    Input: l         galactic longitude in radians
           b         galactic latitude in radians
           d_target  input distance to integrate to (kpc)
           ds_coarse coarse step size along LoS (kpc)
           ds_fine   fine step size along LoS (kpc)
           Nsmin     minimum number of samples to use along LoS
           d2dm_only True => calculate DM only; otherwise also calculate SM.
           do_analysis True => analyze components of line of sight: dm, lism, arms (sm)
           plotting  True => plot dm etc vs distance along path
           verbose   N/A presently

    Output:
           dist          calculated distance or input distance
           dm            calculated DM 
           sm            scattering measure, uniform weighting) (kpc/m^{20/3} 
           smtau         scattering measure, weighting for pulse broadening 
           smtheta       scattering measure, weighting for angular broadening
                          of galactic sources 
           smiso          scattering measure appropriate for calculating the
                           isoplanatic angle at the source's location
           Uses Kolmogorov spectral index = 11/3
           Useful constants: c_sm = (sikol - 3) / (2 * (2*pi)**(4-sikol)) 
                             sm_factor = c_sm * units_conversion to kpc m^{-20/3} 
                             (defined in config_ne2001p.py)       
    """
    # ----------------------------------------------------------------------

    sm_iso_index = sikol - 2                # should be 5/3 for Kolmogorov index sikol=11/3
    limit=' '                               # placeholder for legacy code

    # Coarse and fine integration grids 
    Ns_coarse = int(d_target / ds_coarse)
    if Ns_coarse < Nsmin:
        Ns_coarse = Nsmin
        ds_coarse = d_target / Ns_coarse

    Ns_fine = int(d_target / ds_fine)
    if Ns_fine < Nsmin:
        Ns_fine = Nsmin
        ds_fine = d_target / Ns_fine

    relevant_clump_indices = relevant_clumps(l, b, d_target, rcmult)

    sc_vec, xc_vec, yc_vec, zc_vec = calc_galcentric_vecs(l, b, d_target, Ns_coarse)
    sf_vec, xf_vec, yf_vec, zf_vec = calc_galcentric_vecs(l, b, d_target, Ns_fine)

    # ---------------------------------------------
    # Smooth, large-scale components on coarse grid
    # ---------------------------------------------
    # Note only cnd_smooth, cFsmooth needed here:
    cne1,cne2,cnea, cF1, cF2, cFa, cwhicharm, cne_smooth, cFsmooth = \
        array([
           density_2001_smooth_comps(xc_vec[j],yc_vec[j],zc_vec[j]) for j in range(Ns_coarse)
        ]).T

    # Spline functions:
    cs_ne_smooth = CubicSpline(sc_vec, cne_smooth)
    cs_F_smooth = CubicSpline(sc_vec, cFsmooth)

    # Evaluate using spline functions:
    ne_smooth = cs_ne_smooth(sf_vec)
    F_smooth = cs_F_smooth(sf_vec)

    # Resample cwhicharm using digitize: use coarse vec as bins for fine vec:
    inds_whicharm = np.digitize(sf_vec, sc_vec, right=True)
    whicharm = cwhicharm[inds_whicharm]

    # ------------------------------------
    # Small-scale components on fine grid:
    # ------------------------------------
    # SKO -- tested changing inds_relevant to None from relevant_clump_indices
    negc,nelism,necN,nevN, Fgc, Flism, FcN, FvN, wlism, wldr, wlhb, wlsb, wloopI, \
       hitclump, hitvoid, wvoid = \
           array([
               density_2001_smallscale_comps(\
                   xf_vec[j],yf_vec[j],zf_vec[j], inds_relevant=relevant_clump_indices) \
                   for j in range(Ns_fine)\
           ]).T
    
    wtotal = (1-wgvN*wvoid)*(1-wglism*wlism)        # used for SM calculations
    ne_ex_clumps_voids = (1.-wglism*wlism) * (ne_smooth + wggc*negc) + wglism*wlism*nelism
    ne = (1-wgvN*wvoid)*ne_ex_clumps_voids  + wgvN*wvoid*nevN + wgcN*necN

    dm_cumulate_vec = \
        pc_in_kpc * array([trapz(ne[:j], sf_vec[:j]) for j in range(1, Ns_fine+1) ])
    dm_calc_max = dm_cumulate_vec[-1]       

    # floats -> ints:
    whicharm = whicharm.astype(int)
    hitclump = hitclump.astype(int)
    hitvoid = hitvoid.astype(int)

    if np.count_nonzero(hitclump)!=0:
    	warnings.warn('Clump(s) intersected. Run los_diagnostics.py for details.')
    if np.count_nonzero(hitvoid)!=0:
    	warnings.warn('Void(s) intersected. Run los_diagnostics.py for details.')


    if plotting:
        plot_dm_along_LoS(
            dm_calc_max, d_target, sf_vec, dm_cumulate_vec, relevant_clump_indices, dc, 
            which='d2dm', plot_dm_target=True, saveplot=True)

    if d2dm_only is True:
            if do_analysis:
                if eval_NE2001:
                    savedir = os.getcwd()+'/output_ne2001p/'
                elif eval_NE2025:
                    savedir = os.getcwd()+'/output_ne2025p/'
                f24outfile  = savedir+'f24_d2dm_only_lism_etc_paths' + '.txt'
                f24 = open(f24outfile, 'w')

                f25outfile  = savedir+'f25_d2dm_only_ne_vs_s' + '.txt'
                f25 = open(f25outfile, 'w')

                dhat = d_target             # so that code here is same as in dmdsm_dm2d
                dm_target = dm_calc_max     # ditto
                svec, ne, ne1, ne2, nea = \
                    analysis_dmd_dm_only(f24, f25,
                    l, b, dm_target, dhat, sf_vec, xf_vec, yf_vec, zf_vec, sc_vec,
                    cne1, cne2, cnea, ne, whicharm, hitclump, hitvoid, dm_cumulate_vec,
                    nelism, negc, necN, nevN, wtotal, wlism, wvoid,
                    wlhb, wldr, wlsb, wloopI)
            return limit, d_target, dm_calc_max
    else:

        dhat = d_target             # so that code here is same as in dmdsm_dm2d
        dm_target = dm_calc_max     # ditto

        # Calculate dSM (proportional to Cn2) from individual dSM components
        # Note: dsm1, dsm2, dsma are subsumed by dsm_smooth
        # dsm1, dsm2, dsma are calculated in the `analysis' code block (if requested)

        # For now, the dsm quantities are not multiplied by sm_factor.
        # That is done later in calculating sm quantities from dsm quantities.
        # Could change this to make it a bit more transparent 
        # e.g. multiply dsm quantities by sm_factor here.

        dsm_smooth = wtotal * ne_smooth**2 * F_smooth
        dsmgc = wtotal*wggc*negc**2*Fgc
        dsmlism = (1.-wgvN*wvoid)*wglism*wlism*nelism**2*Flism
        dsmcN = wgcN*necN**2*FcN
        dsmvN = wgvN*wvoid*nevN**2*FvN
        dsm = dsm_smooth + dsmgc + dsmlism + dsmcN + dsmvN

        # Calculate integrals needed to evaluate SM;
        # Note integrals start at observer's position
        # Integrate from s = 0 to s = dhat
        # Accomplish this by integrating over full svec used
        # and then use cubic spline to find SM at d = dhat:.

        Nsf1 = np.size(sf_vec) + 1
        dsm_cumulate1_vec = array([trapz(dsm[:j], sf_vec[:j]) for j in range(1, Nsf1)])
        dsm_cumulate2_vec = \
            array([trapz(sf_vec[:j]*dsm[:j], sf_vec[:j]) for j in range(1, Nsf1)])
        dsm_cumulate3_vec = \
            array([trapz(sf_vec[:j]**2*dsm[:j], sf_vec[:j]) for j in range(1, Nsf1)])
        dsm_cumulate4_vec = \
            array([trapz(sf_vec[:j]**sm_iso_index*dsm[:j], sf_vec[:j]) for j in range(1, Nsf1)])

        # sm quantities have proper SM units:

        sm_cumulate = sm_factor *dsm_cumulate1_vec
        smtau_cumulate = 6 * (sm_factor/dhat) * (dsm_cumulate2_vec - dsm_cumulate3_vec/dhat)
        smtheta_cumulate = 3 * sm_factor * \
            (dsm_cumulate1_vec + dsm_cumulate3_vec/dhat**2 - 2*dsm_cumulate2_vec/dhat)
        smiso_cumulate = sm_factor * dsm_cumulate4_vec

        # cubic splines:

        sm_cs = CubicSpline(sf_vec, sm_cumulate)
        smtau_cs = CubicSpline(sf_vec, smtau_cumulate)
        smtheta_cs = CubicSpline(sf_vec, smtheta_cumulate)
        smiso_cs = CubicSpline(sf_vec, smiso_cumulate)

        sm_hat = sm_cs(dhat)
        smtau_hat = smtau_cs(dhat)
        smtheta_hat = smtheta_cs(dhat)
        smiso_hat = smiso_cs(dhat)

        if plotting:
            plot_dm_sm_along_LoS(
                dm_target, dm_cumulate_vec, dhat, sf_vec,
                sm_hat, smtau_hat, smtheta_hat, smiso_hat,
                dsm, sm_cumulate, smtau_cumulate, smtheta_cumulate, smiso_cumulate,
                which='d2dm', saveplot=True)

        if do_analysis:

            #print(eval_NE2001,eval_NE2025)
            if eval_NE2001:
                savedir = os.getcwd()+'/output_ne2001p/'
            elif eval_NE2025:
                savedir = os.getcwd()+'/output_ne2025p/'

            f24outfile  = savedir+'f24_d2dm_lism_etc_paths' + '.txt'
            f24 = open(f24outfile, 'w')

            f25outfile  = savedir+'f25_d2dm_ne_dsm_vs_s' + '.txt'
            f25 = open(f25outfile, 'w')

            svec, ne, ne1, ne2, nea = \
            analysis_dmd_dm_and_sm(f24, f25, 
                l, b, dm_target, dhat, sf_vec, xf_vec, yf_vec, zf_vec, sc_vec,
                cne1, cF1, cne2, cF2, cnea, cFa,
                sm_hat, ne, dsm, whicharm, hitclump, hitvoid, dm_cumulate_vec,
                nelism, negc, necN, nevN, wtotal, wlism, wvoid,
                dsmgc, dsmlism, dsmcN, dsmvN,
                wlhb, wldr, wlsb, wloopI)

        return limit, dhat, dm_target, sm_hat, smtau_hat, smtheta_hat, smiso_hat

# ----------------------------------------------------------------------

def analysis_dmd_dm_only(f24, f25, 
    l, b, dm_target, dhat, sf_vec, xf_vec, yf_vec, zf_vec, sc_vec,
    cne1, cne2, cnea, ne, whicharm, hitclump, hitvoid, dm_cumulate_vec,
    nelism, negc, necN, nevN, wtotal, wlism, wvoid,
    wlhb, wldr, wlsb, wloopI):
    
        """
        Assess contributions from different model components
        and print to files. 
 
        This version excludes any scattering variables
        (a selection made for faster execution)

        Note that some quantities are defined only here because otherwise
        they would slow down the dm->d calculation.

        Output files are defined outside this function (f24, f25)
        Units:
            l,b                                            rad
            dm_target, dm_cumulate_vec                     pc/cc           
            dhat, sfvec, xfvec, yf_vec, zf_vec, sc_vec     kpc
            cne1, cne2, cnea, ne, nelism, negc, necN, nevN 1/cc
            cF1, cF2, cFa                                  pc^{-2/3}  check

        Dimensionless:
            whicharm, hitclump, hitvoid, wtotal, wlism, wvoid, wlhb, wldr, wlsb, wloopI

        """
        # Calculate smooth components on fine grid using CubicSpline to produce
        # spline functions from coarse samples:

        cs_ne1 = CubicSpline(sc_vec, cne1)
        cs_ne2 = CubicSpline(sc_vec, cne2)
        cs_nea = CubicSpline(sc_vec, cnea)

        # Evaluate using spline functions:
        ne1 = cs_ne1(sf_vec)
        ne2 = cs_ne2(sf_vec)
        nea = cs_nea(sf_vec)

        # print differential values along LoS to file:
        print('DM target: ', dm_target, file=f25)
        print(rad2deg(l), rad2deg(b), ' ldeg, bdeg', file=f25)

        # Entries with '---' are to have placeholders so that output number of fields
        # is the same as for analysis that includes SM.
        print('    d       x       y       z       ne     -------  w    c   v  t     dm       nea     ---', file=f25)

        testdm = where(dm_cumulate_vec <= dm_target, 0, 1)     # 0 if dm < dm_target

        Ns_fine = sf_vec.size
        # print out values only up to one step beyond where sf_vec = dhat:
        inds = where(testdm==0)
        Ns_print = np.min((np.size(sf_vec), np.size(inds)+1))
 
        for n in range(Ns_print):
                    print(
                '{:7.3f}{:8.3f}{:8.3f}{:8.3f}{:9.4f}{:11.4e}{:2d}{:5d}{:4d}{:3b}{:9.3f}{:10.4f}{:8.2f}'.format(sf_vec[n],xf_vec[n],yf_vec[n],zf_vec[n],ne[n],0.,whicharm[n],hitclump[n],hitvoid[n], testdm[n], dm_cumulate_vec[n], nea[n], 0.), file=f25)
        f25.close()

        # Print LoS summary analysis to file 24:

        # --------------------------------------
        # Calculate individual end-point values:
        # Uses CubicSplines
        # --------------------------------------

        ddm1 = wtotal*wg1*ne1 * 1000
        ddm2 = wtotal*wg2*ne2 * 1000
        ddma = wtotal*wga*nea * 1000
        ddmgc = wtotal*wggc*negc * 1000
        ddmlism = (1.-wgvN*wvoid)*wglism*wlism*nelism * 1000
        ddmcN = wgcN*necN * 1000
        ddmvN = wgvN*wvoid*nevN * 1000

        dm1run = array([trapz(ddm1[0:j], sf_vec[0:j]) for j in range(size(ddm1))])
        dm2run = array([trapz(ddm2[0:j], sf_vec[0:j]) for j in range(size(ddm2))])
        dmarun = array([trapz(ddma[0:j], sf_vec[0:j]) for j in range(size(ddma))])
        dmgcrun = array([trapz(ddmgc[0:j], sf_vec[0:j]) for j in range(size(ddmgc))])
        dmlismrun = array([trapz(ddmlism[0:j], sf_vec[0:j]) for j in range(size(ddmlism))])
        dmcNrun = array([trapz(ddmcN[0:j], sf_vec[0:j]) for j in range(size(ddmcN))])
        dmvNrun = array([trapz(ddmvN[0:j], sf_vec[0:j]) for j in range(size(ddmvN))])

        cs_dm1 = CubicSpline(sf_vec, dm1run)
        cs_dm2 = CubicSpline(sf_vec, dm2run)
        cs_dma = CubicSpline(sf_vec, dmarun)
        cs_dmgc = CubicSpline(sf_vec, dmgcrun)
        cs_dmlism = CubicSpline(sf_vec, dmlismrun)
        cs_dmcN = CubicSpline(sf_vec, dmcNrun)
        cs_dmvN = CubicSpline(sf_vec, dmvNrun)

        dm1 = cs_dm1(dhat)
        dm2 = cs_dm2(dhat)
        dma = cs_dma(dhat)
        dmgc = cs_dmgc(dhat)
        dmlism = cs_dmlism(dhat)
        dmcN = cs_dmcN(dhat)
        dmvN = cs_dmvN(dhat)


        #---------------
        # LISM analysis:
        #---------------
        # Calculate path lengths through LISM components:
        # Take into account the weighting hierarchy, LHB:LOOPI:LSB:LDR

        ldr_path = lhb_path = lsb_path = loopI_path = 0.
        ldr_dist = lhb_dist = lsb_dist = loopI_dist = 0.

        lhb_path = trapz(wlhb, sf_vec)
        loopI_path = trapz((1-wlhb) * wloopI, sf_vec)
        lsb_path = trapz((1-wlhb)*(1-wloopI) * wlsb, sf_vec)
        ldr_path = trapz((1-wlhb)*(1-wloopI)*(1-wlsb) * wldr, sf_vec)

        if lhb_path > 0:
            lhb_dist = trapz(wlhb*sf_vec, sf_vec) / lhb_path 
        if loopI_path > 0:
            loopI_dist = trapz((1-wlhb)*wloopI*sf_vec, sf_vec) / loopI_path 
        if lsb_path > 0:
            lsb_dist = trapz((1-wlhb)*(1-wloopI)*wlsb*sf_vec, sf_vec) / lsb_path
        if ldr_path > 0:
            ldr_dist = trapz((1-wlhb)*(1-wloopI)*(1-wlsb)*wldr*sf_vec, sf_vec) / ldr_path

        # --------------------
        # Spiral arm analysis:
        # --------------------
        # Calculate path lengths through spiral arms:
        """
        pathlengths: 
            whicharm = 0,5 (currently).
                       1,4 for the equivalent of the TC93 arms
                         5 for the local arm
                         0 means interarm path segments
        note narmsmax = number of arms
        armpaths has one additional entry to include interarm path length, so
        narmsmax1 = narmsmax + 1 (set in config_ne2001p.py)
        """
        armpaths = zeros(narmsmax1)
        armdistances = zeros(narmsmax1)

        # Using for loop for greater code clarity 
        # (could use list comprehensions but expressions are clunky)
        for j in range(narmsmax1):
            inds = where(whicharm==j) 
            sinds = sf_vec[inds]
            winds = whicharm[inds] / max(j, 1)              # avoid div by zero for j=0
            armpaths[j] = trapz(winds, sinds)
            armdistances[j] = trapz(sinds*winds, sinds) / where(armpaths[j] > 0, armpaths[j], 1) 

        # Printing to f24 file:
        print('LISM path lengths (kpc) with weighting hierarchy LHB:LOOPI:LSB:LDR',
        file = f24)
        print('{}'.format(18*' '+ '  LHB     LoopI     LSB      LDR'.expandtabs(15)),
        file = f24)
        print(3*' '+  '{0:15s}{1:6.3f}{2:9.3f}{3:9.3f}{4:9.3f}'.format('Length', lhb_path, loopI_path, lsb_path, ldr_path), file=f24)
        print(3*' '+  '{0:15s}{1:6.3f}{2:9.3f}{3:9.3f}{4:9.3f}'.format('Mean Dist.', lhb_dist, loopI_dist, lsb_dist, ldr_dist), file=f24)

        print('\nFractional contributions to DM:', file=f24)
        print('  outer   inner    arms     gc    lism    clumps  voids       DM', file= f24)
        print('{:7.3f}{:8.3f}{:8.3f}{:8.3f}{:8.3f}{:8.3f}{:8.3f}{:11.3f}'.format(
            dm1/dm_target, dm2/dm_target, dma/dm_target, dmgc/dm_target,
            dmlism/dm_target, dmcN/dm_target, dmvN/dm_target, dm_target), file=f24)

        print('\nPath lengths through spiral arms:', file=f24)
        print('  Arm      Mean Distance       Path Length    (arm=0 => interarm)', file=f24)

        for j in range(narmsmax1):
           print('  {:2d}{:18.3f}{:18.3f}'.format(j, armdistances[j], armpaths[j]), file=f24)

        f24.close()
        return sf_vec, ne, ne1, ne2, nea


# ----------------------------------------------------------------------

def analysis_dmd_dm_and_sm(f24, f25,
    l, b, dm_target, dhat, sf_vec, xf_vec, yf_vec, zf_vec, sc_vec,
    cne1, cF1, cne2, cF2, cnea, cFa,
    sm, ne, dsm, whicharm, hitclump, hitvoid, dm_cumulate_vec,
    nelism, negc, necN, nevN, wtotal, wlism, wvoid,
    dsmgc, dsmlism, dsmcN, dsmvN,
    wlhb, wldr, wlsb, wloopI):
    
    """
    Assess contributions from different model components
    and print to files. 

    Note that some quantities are defined only here because otherwise
    they would slow down the dm->d calculation.

    e.g. dsm1, dsm2, dsma

    Output files are defined outside this function (f24, f25)

    Units:
        l,b                                            rad
        dm_target, dm_cumulate_vec                     pc/cc
        dhat, sfvec, xfvec, yf_vec, zf_vec, sc_vec     kpc
        cne1, cne2, cnea, ne, nelism, negc, necN, nevN 1/cc
        cF1, cF2, cFa                                  pc^{-2/3}  check
        sm                                             kpc m^{-20/3}
        dsm, dsmgc, dsmlism, dsmcN, dsmvN              (sm units) / sm_factor

        Dimensionless: 
            whicharm, hitclump, hitvoid, wtotal, wlism, wvoid, wlhb, wldr, wlsb, wloopI
    """

    # Calculate smooth components on fine grid using CubicSpline to produce
    # spline functions from coarse samples:

    cs_ne1 = CubicSpline(sc_vec, cne1)
    cs_F1 = CubicSpline(sc_vec, cF1)
    cs_ne2 = CubicSpline(sc_vec, cne2)
    cs_F2 = CubicSpline(sc_vec, cF2)
    cs_nea = CubicSpline(sc_vec, cnea)
    cs_Fa = CubicSpline(sc_vec, cFa)

    # Evaluate using spline functions:
    ne1 = cs_ne1(sf_vec)
    ne2 = cs_ne2(sf_vec)
    nea = cs_nea(sf_vec)
    F1 = cs_F1(sf_vec)
    F2 = cs_F2(sf_vec)
    Fa = cs_Fa(sf_vec)

    # print differential values along LoS to file:
    print('DM target: ', dm_target, file=f25)
    print(rad2deg(l), rad2deg(b), ' ldeg, bdeg', file=f25)

    print('    d       x       y       z       ne         Cn2     w    c   v  t     dm       nea      Fa', file=f25)

    testdm = where(dm_cumulate_vec <= dm_target, 0, 1)     # 0 if dm < dm_target

    Ns_fine = sf_vec.size
    # print out values only up to one beyond where sf_vec = dhat:
    inds = where(testdm==0)
    Ns_print = np.min((np.size(sf_vec), np.size(inds)+1))

    for n in range(Ns_print):
        print(
                '{:7.3f}{:8.3f}{:8.3f}{:8.3f}{:9.4f}{:14.4e}{:2d}{:5d}{:4d}{:3b}{:9.3f}{:10.4f}{:8.2f}'.format(sf_vec[n], xf_vec[n], yf_vec[n], zf_vec[n], ne[n], sm_factor*dsm[n], whicharm[n], hitclump[n], hitvoid[n], testdm[n], dm_cumulate_vec[n], nea[n], Fa[n]), file=f25)
    f25.close()

    # print LoS summary analysis to file

    # --------------------------------------
    # Calculate individual end-point values:
    # Uses CubicSplines
    # --------------------------------------

        # ddm/ds components at fine resolution:
    ddm1 = wtotal*wg1*ne1 * 1000
    ddm2 = wtotal*wg2*ne2 * 1000
    ddma = wtotal*wga*nea * 1000
    ddmgc = wtotal*wggc*negc * 1000
    ddmlism = (1.-wgvN*wvoid)*wglism*wlism*nelism * 1000
    ddmcN = wgcN*necN * 1000
    ddmvN = wgvN*wvoid*nevN * 1000

    # cumulative integrals of n_e components:
    dm1run = array([trapz(ddm1[0:j], sf_vec[0:j]) for j in range(size(ddm1))])
    dm2run = array([trapz(ddm2[0:j], sf_vec[0:j]) for j in range(size(ddm2))])
    dmarun = array([trapz(ddma[0:j], sf_vec[0:j]) for j in range(size(ddma))])
    dmgcrun = array([trapz(ddmgc[0:j], sf_vec[0:j]) for j in range(size(ddmgc))])
    dmlismrun = array([trapz(ddmlism[0:j], sf_vec[0:j]) for j in range(size(ddmlism))])
    dmcNrun = array([trapz(ddmcN[0:j], sf_vec[0:j]) for j in range(size(ddmcN))])
    dmvNrun = array([trapz(ddmvN[0:j], sf_vec[0:j]) for j in range(size(ddmvN))])


    # spline functions for cumulative DM components:
    cs_dm1 = CubicSpline(sf_vec, dm1run)
    cs_dm2 = CubicSpline(sf_vec, dm2run)
    cs_dma = CubicSpline(sf_vec, dmarun)
    cs_dmgc = CubicSpline(sf_vec, dmgcrun)
    cs_dmlism = CubicSpline(sf_vec, dmlismrun)
    cs_dmcN = CubicSpline(sf_vec, dmcNrun)
    cs_dmvN = CubicSpline(sf_vec, dmvNrun)

    # DM component values at estimated distance
    dm1 = cs_dm1(dhat)
    dm2 = cs_dm2(dhat)
    dma = cs_dma(dhat)
    dmgc = cs_dmgc(dhat)
    dmlism = cs_dmlism(dhat)
    dmcN = cs_dmcN(dhat)
    dmvN = cs_dmvN(dhat)

    #-----------------------------
    # Individual components of SM:
        # [Note these need to be multiplied by
        # sm_factor = 1.836 to get SM in kpc m^{-20/3}]
    #-----------------------------

    dsm1 = wtotal*wg1*ne1**2*F1
    dsm2 = wtotal*wg2*ne2**2*F2
    dsma = wtotal*wga*nea**2*Fa

    # cumulative integrals of SM components (still divided by sm_factor)
    sm1run = array([trapz(dsm1[0:j], sf_vec[0:j]) for j in range(size(dsm1))])
    sm2run = array([trapz(dsm2[0:j], sf_vec[0:j]) for j in range(size(dsm2))])
    smarun = array([trapz(dsma[0:j], sf_vec[0:j]) for j in range(size(dsma))])
    smgcrun = array([trapz(dsmgc[0:j], sf_vec[0:j]) for j in range(size(dsmgc))])
    smlismrun = array([trapz(dsmlism[0:j], sf_vec[0:j]) for j in range(size(dsmlism))])
    smcNrun = array([trapz(dsmcN[0:j], sf_vec[0:j]) for j in range(size(dsmcN))])
    smvNrun = array([trapz(dsmvN[0:j], sf_vec[0:j]) for j in range(size(dsmvN))])

    # spline functions for cumulative SM components:
    cs_sm1 = CubicSpline(sf_vec, sm1run)
    cs_sm2 = CubicSpline(sf_vec, sm2run)
    cs_sma = CubicSpline(sf_vec, smarun)
    cs_smgc = CubicSpline(sf_vec, smgcrun)
    cs_smlism = CubicSpline(sf_vec, smlismrun)
    cs_smcN = CubicSpline(sf_vec, smcNrun)
    cs_smvN = CubicSpline(sf_vec, smvNrun)

    # SM components at estimated distance (now with sm_factor):
    sm1 = cs_sm1(dhat) * sm_factor
    sm2 = cs_sm2(dhat) * sm_factor
    sma = cs_sma(dhat) * sm_factor
    smgc = cs_smgc(dhat) * sm_factor
    smlism = cs_smlism(dhat) * sm_factor
    smcN = cs_smcN(dhat) * sm_factor
    smvN = cs_smvN(dhat) * sm_factor

    #---------------
    # LISM analysis:
    #---------------
    # Calculate path lengths through LISM components:
    # Take into account the weighting hierarchy, LHB:LOOPI:LSB:LDR

    ldr_path = lhb_path = lsb_path = loopI_path = 0.
    ldr_dist = lhb_dist = lsb_dist = loopI_dist = 0.

    lhb_path = trapz(wlhb, sf_vec)
    loopI_path = trapz((1-wlhb) * wloopI, sf_vec)
    lsb_path = trapz((1-wlhb)*(1-wloopI) * wlsb, sf_vec)
    ldr_path = trapz((1-wlhb)*(1-wloopI)*(1-wlsb) * wldr, sf_vec)

    if lhb_path > 0:
        lhb_dist = trapz(wlhb*sf_vec, sf_vec) / lhb_path 
    if loopI_path > 0:
        loopI_dist = trapz((1-wlhb)*wloopI*sf_vec, sf_vec) / loopI_path 
    if lsb_path > 0:
        lsb_dist = trapz((1-wlhb)*(1-wloopI)*wlsb*sf_vec, sf_vec) / lsb_path
    if ldr_path > 0:
        ldr_dist = trapz((1-wlhb)*(1-wloopI)*(1-wlsb)*wldr*sf_vec, sf_vec) / ldr_path

    # --------------------
    # Spiral arm analysis:
    # --------------------
    # Calculate path lengths through spiral arms:
    """
    pathlengths: 
        whicharm = 0,5 (currently).
            1,4 for the equivalent of the TC93 arms
            5 for the local arm
            0 means interarm path segments
    note narmsmax = number of arms
    armpaths has one additional entry to include interarm path length, so
    narmsmax1 = narmsmax + 1 (set in config_ne2001p.py)
    """
    armpaths = zeros(narmsmax1)
    armdistances = zeros(narmsmax1)

    # Using for loop for greater code clarity 
    # (could use list comprehensions but expressions are clunky)
    for j in range(narmsmax1):
        inds = where(whicharm==j) 
        sinds = sf_vec[inds]
        winds = whicharm[inds] / max(j, 1)              # avoid div by zero for j=0
        armpaths[j] = trapz(winds, sinds)
        armdistances[j] = trapz(sinds*winds, sinds) / where(armpaths[j] > 0, armpaths[j], 1) 

    # Printing to f24 file:
    print('LISM path lengths (kpc) with weighting hierarchy LHB:LOOPI:LSB:LDR',
    file = f24)
    print('{}'.format(18*' '+ '  LHB     LoopI     LSB      LDR'.expandtabs(15)),
    file = f24)
    print(3*' '+  '{0:15s}{1:6.3f}{2:9.3f}{3:9.3f}{4:9.3f}'.format('Length', lhb_path, loopI_path, lsb_path, ldr_path), file=f24)
    print(3*' '+  '{0:15s}{1:6.3f}{2:9.3f}{3:9.3f}{4:9.3f}'.format('Mean Dist.', lhb_dist, loopI_dist, lsb_dist, ldr_dist), file=f24)

    print('\nFractional contributions to DM:', file=f24)
    print('  outer   inner    arms     gc    lism    clumps  voids       DM', file= f24)
    print('{:7.3f}{:8.3f}{:8.3f}{:8.3f}{:8.3f}{:8.3f}{:8.3f}{:11.3f}'.format(
        dm1/dm_target, dm2/dm_target, dma/dm_target, dmgc/dm_target,
        dmlism/dm_target, dmcN/dm_target, dmvN/dm_target, dm_target), file=f24)

    print('\nFractional contributions to SM:', file=f24)
    print('  outer   inner    arms     gc    lism    clumps  voids       SM', file= f24)
    print('{:7.3f}{:8.3f}{:8.3f}{:8.3f}{:8.3f}{:8.3f}{:8.3f}{:11.3e}'.format(
        sm1/sm, sm2/sm, sma/sm, smgc/sm,
        smlism/sm, smcN/sm, smvN/sm, sm), file=f24)

    print('\nPath lengths through spiral arms:', file=f24)
    print('  Arm      Mean Distance       Path Length    (arm=0 => interarm)', file=f24)

    for j in range(narmsmax1):
         print('  {:2d}{:18.3f}{:18.3f}'.format(j, armdistances[j], armpaths[j]), file=f24)

    f24.close()
    return sf_vec, ne, ne1, ne2, nea

# ----------------------------------------------------------------------

def plot_dm_along_LoS(
        dm_target, dhat, sf_vec, dm_cumulate_vec, relevant_clump_indices, dc, 
        which = 'dm2d', plot_dm_target=False, saveplot=True):

        """
        Plots DM(s) along line of sight

        which = 'dm2d' or 'd2dm': controls labeling
        """ 

        # Plot cumulative DM 
        fig = figure()
        ax = fig.add_subplot(111)
        subplots_adjust(bottom=0.15)
        plot(sf_vec, dm_cumulate_vec, label=r'$\rm Model \ DM$')

        # Plot clump DMs if any affect the LoS
        if np.size(relevant_clump_indices) > 0:
            clump_distances = dc[relevant_clump_indices]
            clump_dm_hats = interp(clump_distances, sf_vec, dm_cumulate_vec) 
            plot(clump_distances, clump_dm_hats, 'ro', 
                label=r'$\rm  DM\,@\ clump \ positions)$')
        axis(xmin=-0.05*sf_vec[-1])
        axis(ymin=-0.05*dm_cumulate_vec[-1])
        plot((axis()[0], dhat), (dm_target, dm_target), 'k--', lw=1)
        plot((dhat, dhat), (axis()[2], dm_target), 'k--', lw=1)
        xlabel(r'$\rm Distance \ along \ LoS \ \ \  (kpc)$', fontsize=13)
        ylabel(r'$\rm DM(s) \ \    (pc\ cm^{-3})$', fontsize=13)

        if which == 'dm2d':
            title(r'$\rm DM\,\to\, D: \ \ DM_{target} = %5.1f \ pc\ cm^{-3} \quad\quad\quad \widehat d = %5.2f \ kpc$'%(dm_target, dhat), fontsize=12)
        if which == 'd2dm':
            title(r'$\rm D\,\to\, DM: \ \ D_{target} = %6.2f \ kpc  \quad\quad\quad \widehat DM(d_{\rm target})  = %5.1f \ pc\,cm^{-3}$'%(dhat, dm_cumulate_vec[-1]), fontsize=12)
            if plot_dm_target:
                plot(dhat, dm_target, 'ko', label=r'$\rm Target \ DM$')
        legend(loc=0, fontsize=10)
        annotate(plotstamp, xy=(0.70, 0.02), xycoords='figure fraction', ha='left', va='center', fontsize=5)

        if eval_NE2001:
            savedir = os.getcwd()+'/output_ne2001p/'
        elif eval_NE2025:
            savedir = os.getcwd()+'/output_ne2025p/'

        if saveplot:
            #plotfile = 'dm_vs_d_' + basename + '.pdf'
            plotfile = savedir+'dm_vs_d.pdf'
            savefig(plotfile)
        show()
        
        return

# ----------------------------------------------------------------------

def plot_dm_sm_along_LoS(
    dm_target, dm_cumulate_vec, dhat, sf_vec,
    sm_hat, smtau_hat, smtheta_hat, smiso_hat,
    dsm, sm_cumulate, smtau_cumulate, smtheta_cumulate, smiso_cumulate,
    which = 'dm2d', saveplot=True):
    """
    Plots DM(s) and SM(s) along line of sight
    """

    # Plot cumulative DM and SMs
    fig = figure(figsize=(6,6))
    ax = fig.add_subplot(311)
    plot(sf_vec, sm_factor * dsm)
    ylabel(r'$\rm \Delta SM$',fontsize=13)
    if which == 'dm2d':
        title(r'$\rm DM\to D: \ \ \hat d = %5.2f \ \ SM, SM_\tau, SM_\theta = %8.4f \  %8.4f \ %8.4f$' %(dhat, sm_hat, smtau_hat, smtheta_hat), fontsize=10) 
    if which == 'd2ddm':
        title(r'$\rm D\to DM: \ \ \hat d = %5.2f \ \ SM, SM_\tau, SM_\theta = %8.4f \  %8.4f \ %8.4f$' %(dhat, sm_hat, smtau_hat, smtheta_hat), fontsize=10) 

    ax = fig.add_subplot(312)
    plot(sf_vec, sm_cumulate, label='SM')
    plot(sf_vec, smtau_cumulate, label=r'SM$_{\tau}$')
    plot(sf_vec, smtheta_cumulate, label=r'SM$_{\theta}$')
    plot(sf_vec, smiso_cumulate, label=r'SM$_{\rm iso}$')

    plot(dhat, sm_hat, 'o')
    plot(dhat, smtau_hat, 'o')
    plot(dhat, smtheta_hat, 'o')
    plot(dhat, smiso_hat, 'o')
    legend(loc=0)
    ylabel(r'$\rm SM(s)$',fontsize=13)

    ax = fig.add_subplot(313)
    plot(sf_vec, dm_cumulate_vec)
    xlabel(r'$\rm Distance \ along \ LoS \ \   (kpc)$', fontsize=13)
    ylabel(r'$\rm DM(s) \ \    (pc\ cm^{-3})$', fontsize=13)
    annotate(plotstamp, xy=(0.70, 0.02), xycoords='figure fraction', ha='left', va='center', fontsize=5)
    #show()

    fig.tight_layout()

    if eval_NE2001:
        savedir = os.getcwd()+'/output_ne2001p/'
    elif eval_NE2025:
        savedir = os.getcwd()+'/output_ne2025p/'
    
    if saveplot:
        plotfile = 'sm_and_dm_d.pdf'
        savefig(savedir+plotfile)
    show()

# ----------------------------------------------------------------------

# Main

if __name__ == '__main__':
    
    ldeg, bdeg, dm_target = 30, 0, 1000
    ndir = -1
    ds_fine = 0.005 # 0.01
    ds_coarse = 0.1

    l = deg2rad(ldeg)
    b = deg2rad(bdeg)

    verbose = True
    do_analysis = True
    dm2d_only = True
    dm2d_only = False
    plotting = False
    plotting = True
    debug = False 


    if dm2d_only:
        limit, dhat, dm_target \
        = dmdsm_dm2d(l, b, dm_target, ds_fine=ds_fine, ds_coarse=ds_coarse, 
              Nsmin=10, dm2d_only=dm2d_only, do_analysis=do_analysis, 
              plotting=plotting, verbose=verbose, debug=debug)
    else:
        limit, dhat, dm_target, sm, smtau, smtheta, smiso \
        = dmdsm_dm2d(l, b, dm_target, ds_fine=ds_fine, ds_coarse=ds_coarse, 
            Nsmin=10, dm2d_only=dm2d_only, do_analysis=do_analysis, 
            plotting=plotting, verbose=verbose, debug=debug)

    if plotting:
        input('hit return')
        close('all')




#!/usr/bin/env python

# mwprop v2.0 Jan 2026

'''

tests NE2025p runs properly and gives expected output

'''

from mwprop.nemod import *
from mwprop.nemod.NE2025 import ne2025


if __name__ == '__main__':


	# J0323+3944, D_obs = 0.95, DM_obs = 26.19
	test_gl = 152.180
	test_gb = -14.338
	test_dm_from_d = 16.2718
	test_dkpc_from_dm = 1.3863
	test_sm_from_d = 1.3794e-04
	test_sm_from_dm = 3.5591e-04

	Dk1,Dv1,Du1,Dd1 = ne2025(test_gl,test_gb,0.95,-1,classic=False,dmd_only=False)
	Dk2,Dv2,Du2,Dd2 = ne2025(test_gl,test_gb,26.19,1,classic=False,dmd_only=False)

	print('Testing NE2025p integrations give expected outputs ... Percent errors listed below (values should be less than a few percent):')

	try:
		DM1 = Dv1['DM']
		SM1 = Dv1['SM']
		DM_err = 100*abs(DM1-test_dm_from_d)/test_dm_from_d
		SM_err = 100*abs(SM1-test_sm_from_d)/test_sm_from_d
		print('D->DM: ', DM_err)
		print('D->SM: ', SM_err)

	except:
		print('Error: cannot find expected output (D->DM)')

	try:
		DIST2 = Dv2['DIST']
		SM2 = Dv2['SM']
		Derr = 100*abs(DIST2-test_dkpc_from_dm)/test_dkpc_from_dm
		SMerr = 100*abs(SM2-test_sm_from_dm)/test_sm_from_dm
		print('DM->D: ', Derr)
		print('DM->SM: ', SM_err)

	except:
		print('Error: cannot find expected output (DM->D)')






# mwprop v2.0 Jan 2026

'''
Pythonic versions of the following functions in density.NE2001.f:

ne_outer
ne_inner
ne_gc

Change Log:

01/23/20 -- SKO
01/29/20 -- JMC 
    * added return quantities to ne functions 
    * moved input parameter read to master script
02/08/20 -- JMC
    * replaced parameter input read with import of config_ne2001p
    * rsun = 8.5 now commented out; set in config file
    * other parameters and function defs also now in config file
'''


# nemod_config sets up the model with all dictionaries etc. 
from mwprop.nemod.config_nemod import *

pihalf = np.pi/2.
sqrt = np.sqrt

def ne_outer(x,y,z): #thick disk component

    #g1 = sech2(rr/A1)/sech2(rsun/A1) #TC93 function
    rr = sqrt(x**2 + y**2)
    suncos = cos(pihalf*rsun/A1)
    if rr > A1:
        g1 = 0.
    else:
        g1 = cos(pihalf*rr/A1)/suncos
    ne1 = (n1h1/h1)*g1*sech2(z/h1)
    ne_outer = ne1
    F_outer = F1
    
    return ne_outer, F_outer

def ne_inner(x,y,z): #thin disk component

    g2 = 0.
    rr = sqrt(x**2. + y**2.)
    rrarg = ((rr-A2)/1.8)**2.
    if rrarg < 10.:
        g2 = exp(-rrarg)
    ne2 = n2*g2*sech2(z/h2)
    ne_inner = ne2
    F_inner = F2
    
    return ne_inner, F_inner

def ne_gc(x,y,z, absymax=2*rgc):

    '''
    Determines the contribution of the Galactic center to the free
    electron density of the interstellar medium at Galactic location
    (x,y,z).  Combine with `fluctuation' parameter to obtain the
    scattering measure.
    
    NOTE: This is for the hyperstrong scattering region in the
    Galactic center.  It is distinct from the inner Galaxy
    (component 2) of the TC93 model.
    
    Origin of coordinate system is at Galactic center; the Sun is at
    (x,y,z) = (0,rsun,0), x is in l=90 direction
    
    Based on Section 4.3 of Lazio & Cordes (1998, ApJ, 505, 715)
    
    Input:
    x - location in Galaxy [kpc]
    y - location in Galaxy [kpc]
    z - location in Galaxy [kpc]
    
    COMMON:
    NEGC0 - nominal central density
    
    PARAMETERS:
    RGC - radial scale length of Galactic center density enhancement
    HGC - z scale height of Galactic center density enhancement
    
    Output:
    NE_GC - Galactic center free electron density contribution [cm^-3]

    Definitions now in config_ne2001p.py
    xgc = Dgc['xgc']
    ygc = Dgc['ygc']
    zgc = Dgc['zgc']
    rgc = Dgc['rgc']
    hgc = Dgc['hgc']
    negc0 = Dgc['negc0']
    Fgc0 = Dgc['Fgc0']
    
    '''

    # GC component is nonzero only for abs(y) < rgc (currently)
    # so conservatively exit function for abs(y) > absymax = multiple of rgc:

    if abs(y) > absymax:               
        return 0, 0
    
    rr = sqrt((x-xgc)**2. + (y-ygc)**2.)        #galactocentric radius
    zz = abs(z-zgc)                             #z-height

    if rr > rgc or abs(z-zgc) > hgc:
        return 0, 0
    else:
        arg = (rr/rgc)**2. + (zz/hgc)**2.
        if arg <= 1:                        
            ne_gc_out = negc0
            F_gc = Fgc0
        else: # need this to avoid errors when arg is not <=1
            ne_gc_out = 0
            F_gc = 0

    return ne_gc_out, F_gc



# mwprop.nemod v2.0 Jan 2026

'''
Configuration file for NE20x models
Reads input files and puts in dictionaries
Sets up spiral arms and also puts in dictionaries
'''

global rsun, rf_ref, vperp, sikol, louter, linner, ds_fine, ds_coarse
global Dgal, Dgc, Dlism, Dclumps, Dvoids, Darms, Darmmap, armmap, \
           r1, th1, th1deg, coarse_arms, rarray, tharray, \
           armsplines, armarray, tangents, normals, curvatures

global wg1, wg2, wga, wggc, wglism, wgcN, wgvN
global n1h1, h1, A1, F1, n2, h2, A2, F2, na, ha, wa, Aa, Fa
global apldr, bpldr, cpldr, dpldr, yldrmin, yldrmax
global aplsb, bplsb, cplsb, dplsb, ylsbmin, lsbmax
global alhb, blhb, clhb, xlhb, ylhb, zlhb, thetalhb, nelhb0, Flhb, ylhbmin, ylhbmax
global xlpI, ylpI, zlpI, rlpI, drlpI, nelpI, dnelpI, FlpI, dFlpI
global y_lism_min, y_lism_max
global xgc, ygc, zpgc, rgc, hgc, negc0, Fgc0
global eval_NE2025,eval_NE2001

import numpy as np
from scipy.interpolate import CubicSpline
from scipy.interpolate import UnivariateSpline
from scipy.optimize import fsolve, root, brentq
from scipy.optimize import minimize_scalar
import mpmath as mp

from matplotlib.pyplot import figure, subplots_adjust, plot, axis 
from matplotlib.pyplot import xlabel, ylabel, title, annotate
from matplotlib.pyplot import tick_params, legend, show, close, savefig

import os
import sys
import argparse

script_path = os.path.dirname(os.path.realpath(__file__))

from mwprop.get_constants_waveprop_scipy_version import *

# Get parameter input 
from mwprop.nemod import ne_input

# get some additional units from astropy
from astropy import units as u

# number of pc in kpc:
pc_in_kpc = (1* u.kpc / u.pc).cgs.value

sin = np.sin
cos = np.cos
exp = np.exp
sqrt = np.sqrt
zeros = np.zeros

deg2rad = np.deg2rad
rad2deg = np.rad2deg
sech2 = lambda z: mp.sech(z)**2


# Solar system to Galactic center distance (kpc) set here
rsun = 8.5                              

# Reference radio frequency for calculating chromatic quantities
rf_ref = 1.             # GHz

# Default transverse velocity for calculating scintillation time scale
vperp = 100             # km/s

# Default spectral index for electron density wavenumber spectrum
sikol = 11/3                                            # spectral index 

# Reference outer scale for Kolmogorov spectrum
louter = 1              # pc

# Reference inner scale for Kolmogorov spectrum
linner = 1000           # km

# coarse and fine sampling intervals for LoS numerical integrations
#ds_fine = 0.005 # SKO 1/9/2026 -- 0.0025 leads to spurious clump intersection for J1642, 0.01 made the discrepancy worse!
ds_fine = 0.0025 # SKO 3/2/22 -- changed from 0.01, reduces discrepancies from Fortran version
ds_coarse = 0.1

def setup_spiral_arms(Ncoarse=20, narmpoints=500, drfine=0.01):
    """
    Sets up spiral arm arrays to be done once per call to NE20x package.

    *** maybe rename to 'setup model' and put in higher-level routine

    Input:
        Ncoarse = number of coarse samples along spiral arms
        drfine = sample interval in radius for finely sampled arms
    """

    if 'Darms' not in dir():

        # Obtain model dictionaries and put spiral arm values in arrays

        #print('Creating model dictionaries')
        Dgal, Dgc, Dlism, Dclumps, Dvoids, Darms, eval_NE2025, eval_NE2001 = \
            ne_input.read_nemod_parameters()

        aarm = Darms['a']
        rmin = Darms['rmin']
        thmin = Darms['thmin']
        extent = Darms['extent']
        narms = aarm.size

        # armmap maps integer order  to TC93 order,
        # which is from GC outwards toward Sun.

        armmap = np.array([1, 3, 4, 2, 5])
        Darmmap = {}
        for j in range(narms): 
            Darmmap[str(j)] = armmap[j]

        """
        First evaluate spiral arms on a coarse grid of size Ncoarse
        This is used to find the nearest arm point on a coarse grid
        before calculating to higher precision
        """

        th1 = zeros((narms, Ncoarse))
        r1 = zeros((narms, Ncoarse))

        coarse_arms = zeros((2, narms, Ncoarse))

        extarray = np.linspace(zeros(5), extent, Ncoarse, endpoint=True).T  # rad
        th1 = np.array([thmin[j] + extarray[j] for j in range(narms)])         # rad
        th1deg = np.rad2deg(th1)                                            # deg
        r1 = np.array([rmin[j] * np.exp(extarray[j]/aarm[j]) for j in range(narms)])

        # Sculpt spiral arms as in NE2001 fortran code:

        ### NOTE ARM NUMBERING DIFFERENT FROM TC93 ARM NUMBERS
        ### See Fortran code

        # armindex = 1 ==> Arm 2 = TC arm 3
        inds2a = np.where((370 < th1deg[1]) & (th1deg[1] <= 410))
        arg2a = deg2rad((th1deg[1,inds2a]-390)*180/40.)
        r1[1, inds2a] *= (1 + 0.04*np.cos(arg2a))

        inds2b = np.where((315 < th1deg[1]) & (th1deg[1] <= 370))
        arg2b = deg2rad((th1deg[1,inds2b]-345)*180/55.)
        r1[1, inds2b] *= (1 - 0.07*np.cos(arg2b))

        inds2c = np.where((180 < th1deg[1]) & (th1deg[1] <= 315))
        arg2c = deg2rad((th1deg[1,inds2c]-260)*180/135.)
        r1[1, inds2c] *= (1 + 0.16*np.cos(arg2c))

        # armindex = 3 ==> Arm 4 = TC arm 2
        inds4a = np.where((290 < th1deg[3]) & (th1deg[3] <= 395))
        arg4a = deg2rad((th1deg[3,inds4a]-350)*180/105)
        r1[3, inds4a] *= (1 - 0.11*np.cos(arg4a))

        # r1 is now no longer monotonic splines need as independent variable

        coarse_arm_x = -r1 * np.sin(th1)
        coarse_arm_y =  r1 * np.cos(th1)

        coarse_arms[0]  = coarse_arm_x
        coarse_arms[1]  = coarse_arm_y

        # full spline for plotting outside routine

        tangents = zeros((2, narms, narmpoints))
        normals = zeros((2, narms, narmpoints))
        curvatures = zeros((narms, narmpoints))

        armarray = zeros((2, narms, narmpoints))
        armsplines = []

        rarray = zeros((narms, narmpoints))
        tharray = zeros((narms, narmpoints))
        for j in range(narms):
            sj = CubicSpline(th1[j,:], r1[j,:])
            rj = np.linspace(th1[j,0], th1[j,-1], num=narmpoints, endpoint=True)
            thj = np.linspace(th1[j,0], th1[j,-1], num=narmpoints, endpoint=True)
            rj = sj(thj)
            rarray[j] = rj
            tharray[j] = thj

            # Derivatives
            dthdr = 1 / sj(thj, 1)
            d2thdr2 = 1 / sj(thj, 2)

            dxdr = -np.sin(thj) - rj*np.cos(thj)*dthdr
            dydr =  np.cos(thj) - rj*np.sin(thj)*dthdr
            dydx = dydr / dxdr

            d2xdr2 = \
              -np.cos(thj)*dthdr + rj*np.sin(thj)*dthdr**2 - rj*np.cos(thj)*d2thdr2
            d2ydr2 = \
              -np.sin(thj)*dthdr - rj*np.cos(thj)*dthdr**2 - rj*np.sin(thj)*d2thdr2
            d2ydx2 = d2ydr2 / dxdr**2 + 1. / (d2xdr2 * dxdr * dydr)

            # unit tangent vector:
            facvec = np.array([1./sqrt(1.+dydx[k]**2) for k in range(np.size(dydx))])
            tx = facvec
            ty = facvec * dydx
            tangents[0, j] = tx
            tangents[1, j] = ty

            # unit normal vector:
            nx = -ty
            ny = tx


            normals[0, j] = nx
            normals[1, j] = ny

            Kr = (dxdr*d2ydr2 - dydr*d2xdr2) / (dxdr**2 + dydr**2)**1.5 # Curvature
            curvatures[j] = Kr

            armsplines = np.append(armsplines, sj)
            armarray[0, j, :] = -rj * np.sin(thj)
            armarray[1, j, :] =  rj * np.cos(thj)

    return Dgal, Dgc, Dlism, Dclumps, Dvoids, Darms, Darmmap, \
           armmap, \
           r1, th1, th1deg, coarse_arms, rarray, tharray, \
           armsplines, armarray, tangents, normals, curvatures, eval_NE2025, eval_NE2001


# General parameters for model

"""
Values and coefficients for a 3D Kolmogorov spectrum

For SM modeling (as in NE2001): 
Units conversion for SM in kpc m^{-20/3} and for outer scale in pc
i.e. SM \propto distance x l_o^{-2/3} x n_e^2
so with distance in kpc,  l_o in pc, and n_e in cm^{-3} we get
(kpc in cm) x (l_o in cm)^{-2/3} x (n_e in cm^{-3})^2 = 1.456e9 cm^{-17/3}

Then we divide by (kpc in cm) to get kpc cm^{-20/3}.
Finally converting to m^{-20/3} we have
c_u = (l_o in cm)^{-2/3} x (n_e in cm^{-3})^2 x (cm per m)^{20/3}
    = pc^{-2/3) x 10^{40/3}
    = (3.086e18)^{-2/3} x 10^{40/3}
    \simeq 10.165

For straight units conversion from kpc m^{-20/3} to cm^{-17/3}
we have 
     SMunit (cgs)  = kpc*10.**(-40./3.)             # kpc m^-20/3 to cgs
defined in get_constants_waveprop_astropy.py (and scipy version)
We don't use this in the NE2001 modeling code.. 
"""
c_sm = (sikol-3) / (2 * (2*pi)**(4-sikol))              # ~ 0.181
c_u_units = ((1 * u.m)**(20/3) / (u.pc)**(2/3))         # using astropy units
c_u_astropy = c_u_units.cgs.value                       # evaluate astropy units
sm_factor = c_sm * c_u_astropy                          # net coefficient

dmax = 50                   # Maximum distance returned   (kpc)     need to rename in code
narmsmax = 5                # Number of spiral arms
narmsmax1 = narmsmax + 1

# New for NE2001p
dmax_ne2001p_integrate = 50 # maximum distance to integrate to for 'disk' components

"""
Ncoarse is the number of samples the log-spirals are initially evaluated
Ncoarse  = 20    # as in Fortran code; produces discontinuities
Ncoarse = 100    # reduces discontinuities, but no better than Ncoarse=50
"""

Ncoarse = 50     # reduces discontinuities

Dgal, Dgc, Dlism, Dclumps, Dvoids, Darms, Darmmap, \
    armmap, \
    r1, th1, th1deg, coarse_arms, rarray, tharray,  armsplines, armarray, \
    tangents, normals, curvatures, eval_NE2025, eval_NE2001 = setup_spiral_arms(Ncoarse=Ncoarse)

# ------------------------------
# Weights of density components:
# ------------------------------
wg1 = Dgal['wg1']
wg2 = Dgal['wg2']
wga = Dgal['wga']
wggc = Dgal['wggc']
wglism = Dgal['wglism']
wgcN = Dgal['wgcN']
wgvN = Dgal['wgvN']

# --------------------------
# Main component parameters:
# --------------------------
n1h1 = Dgal['n1h1']
h1 = Dgal['h1']
A1 = Dgal['A1']
F1 = Dgal['F1']
n2 = Dgal['n2']
h2 = Dgal['h2']
A2 = Dgal['A2']
F2 = Dgal['F2']
na = Dgal['na']
ha = Dgal['ha']
wa = Dgal['wa']
Aa = Dgal['Aa']
Fa = Dgal['Fa']


# ---------------------
# Local ISM Components:
# ---------------------
# Note theta values are converted to radians here

# ----
# LDR:
# ----
aldr,bldr,cldr = [Dlism[x] for x in ['aldr', 'bldr', 'cldr']]
xldr,yldr,zldr = [Dlism[x] for x in ['xldr','yldr','zldr']]
thetaldr,neldr0,Fldr = [Dlism[x] for x in ['thetaldr', 'neldr', 'Fldr']]
thetaldr = deg2rad(thetaldr)
sthldr = np.sin(thetaldr)
cthldr = np.cos(thetaldr)

# evaluate quantities used in neLDRQ1
apldr = (cthldr/aldr)**2 + (sthldr/bldr)**2
bpldr = (sthldr/aldr)**2 + (cthldr/bldr)**2
cpldr = 1./cldr**2
dpldr =  2.*cthldr*sthldr*(1./aldr**2 - 1./bldr**2)

yldrmin = yldr-max(aldr,bldr,cldr)
yldrmax = yldr+max(aldr,bldr,cldr)



# ----
# LSB:
# ----
alsb,blsb,clsb = [Dlism[x] for x in ['alsb','blsb','clsb']]
xlsb,ylsb,zlsb =  [Dlism[x] for x in ['xlsb', 'ylsb', 'zlsb']]
thetalsb,nelsb0,Flsb =  [Dlism[x] for x in ['thetalsb', 'nelsb', 'Flsb']]
thetalsb = deg2rad(thetalsb)
sthlsb = np.sin(thetalsb)
cthlsb = np.cos(thetalsb)

# evaluate quantities used in neLSB:
aplsb = (cthlsb/alsb)**2 + (sthlsb/blsb)**2
bplsb = (sthlsb/alsb)**2 + (cthlsb/blsb)**2
cplsb = 1./clsb**2
dplsb =  2.*cthlsb*sthlsb*(1./alsb**2 - 1./blsb**2)

ylsbmin = ylsb-max(alsb,blsb,clsb)
ylsbmax = ylsb+max(alsb,blsb,clsb)


# ----
# LHB:
# ----

alhb,blhb,clhb =  [Dlism[x] for x in ['alhb', 'blhb', 'clhb']]
xlhb,ylhb,zlhb =  [Dlism[x] for x in ['xlhb', 'ylhb', 'zlhb']]
thetalhb,nelhb0,Flhb = [Dlism[x] for x in ['thetalhb', 'nelhb', 'Flhb']]
thetalhb = deg2rad(thetalhb)

ylhbmin = ylhb-max(alhb,blhb,clhb)
ylhbmax = ylhb+max(alhb,blhb,clhb)


xlpI,ylpI,zlpI = [Dlism[x] for x in ['xlpI', 'ylpI', 'zlpI']]
rlpI,drlpI = [Dlism[x] for x in ['rlpI', 'drlpI']]
nelpI,dnelpI,FlpI,dFlpI =  [Dlism[x] for x in ['nelpI','dnelpI','FlpI','dFlpi']]

ylpImin = ylpI - rlpI - drlpI
ylpImax = ylpI + rlpI + drlpI

# y_lism_min = minimum y value for LISM components to matter 
# y_lism_max = maximum y value for LISM components to matter 

y_lism_min = min((yldrmin, ylsbmin, ylhbmin, ylpImin)) 
y_lism_max = max((yldrmax, ylsbmax, ylhbmax, ylpImax)) 



# --------------------------
# Galactic center component:
# --------------------------
xgc = Dgc['xgc']
ygc = Dgc['ygc']
zgc = Dgc['zgc']
rgc = Dgc['rgc']
hgc = Dgc['hgc']
negc0 = Dgc['negc0']
Fgc0 = Dgc['Fgc0']



# NOTE for now no global variables for clumps and voids


# -------
# Clumps:
# -------
nclumps = len(Dclumps)
#print('number of clumps:', nclumps)
lc = np.zeros(len(Dclumps))
bc = np.zeros(len(Dclumps))
nec = np.zeros(len(Dclumps))
Fc = np.zeros(len(Dclumps))
dc = np.zeros(len(Dclumps))
rc = np.zeros(len(Dclumps))
edgec = np.zeros(len(Dclumps))
for n,i in enumerate(Dclumps):
    lc[n] = Dclumps[i]['l']
    bc[n] = Dclumps[i]['b']
    nec[n] = Dclumps[i]['nec']
    Fc[n] = Dclumps[i]['Fc']
    dc[n] = Dclumps[i]['dc']
    rc[n] = Dclumps[i]['rc']
    edgec[n] = Dclumps[i]['edge']

slc = np.sin(deg2rad(lc))
clc = np.cos(deg2rad(lc))
sbc = np.sin(deg2rad(bc))
cbc = np.cos(deg2rad(bc))
rgalc = dc*cbc
xc = rgalc*slc
yc = rsun-rgalc*clc
zc = dc*sbc
# 11/23 SKO -- defining rcmult based on clump edge; this only gets used to filter the clumps, not the integration
rcmult = np.ones(np.shape(edgec))
rcmult[edgec==0] = 2.5 # roll-off
rcmult[edgec==1] = 1.1 # hard cutoff (only for a few clumps, e.g. Gum)

# ------
# Voids:
# ------
nvoids = len(Dvoids)
#print('number of voids:', nvoids)
lv = np.zeros(len(Dvoids))
bv = np.zeros(len(Dvoids))
dv = np.zeros(len(Dvoids))
nev = np.zeros(len(Dvoids))
Fv = np.zeros(len(Dvoids))
aav = np.zeros(len(Dvoids))
bbv = np.zeros(len(Dvoids))
ccv = np.zeros(len(Dvoids))
thvy = np.zeros(len(Dvoids))
thvz = np.zeros(len(Dvoids))
edgev = np.zeros(len(Dvoids))
for n,i in enumerate(Dvoids):
    lv[n] = Dvoids[i]['l']
    bv[n] = Dvoids[i]['b']
    dv[n] = Dvoids[i]['dv']
    nev[n] = Dvoids[i]['nev']
    Fv[n] = Dvoids[i]['Fv']
    aav[n] = Dvoids[i]['aav']
    bbv[n] = Dvoids[i]['bbv']
    ccv[n] = Dvoids[i]['ccv']
    thvy[n] = Dvoids[i]['thvy']
    thvz[n] = Dvoids[i]['thvz']
    edgev[n] = Dvoids[i]['edge']

slv = np.sin(deg2rad(lv))
clv = np.cos(deg2rad(lv))
sbv = np.sin(deg2rad(bv))
cbv = np.cos(deg2rad(bv))
rgalc = dv*cbv
xv = rgalc*slv
yv = rsun - rgalc*clv
zv = dv*sbv
# th1 = thvy
# th2 = thvz
s1 = np.sin(deg2rad(thvy))
c1 = np.cos(deg2rad(thvy))
s2 = np.sin(deg2rad(thvz))
c2 = np.cos(deg2rad(thvz))
cc12 = c1*c2
ss12 = s1*s2
cs21 = c2*s1
cs12 = c1*s2



# mwprop v2.0 Jan 2026

'''
Python version of subroutine nevoidN.f in NE2001 Fortran code 

Returns electron density nevN and fluctuation parameter FvN 
at position designated by l,b,d,x,y,z c for a set of  
voids with parameters read in from file  nevoidN.dat

input:
    x,y,z   coordinates (kpc)  (as in TC93)

output:
    nevN    electron density in void at (x,y,z)
    FvN fluctuation parameter
    hitvoid =   0:     no void hit
                j>0:   j-th void hit
    wvoid = 0,1:     void weight    

parameters:
    lv  = galactic longitude of void center
    bv  = galactic latitude of void center
    dv  = distance from Sun of void center
    (xv,yv,zv) = void center location (calculated)
        nev = internal peak electron density
        Fv      = void fluctuation parameter
    aav = void major axis at 1/e
    bbv = void minor axis at 1/e
    ccv = void minor axis at 1/e
    thvy    = rotation axis of void about y axis
    thvz    = rotation axis of void about z axis
    edgev   = 0 => use exponential rolloff out to 5rc
                 1 => uniform and truncated at 1/e
Version history:

01/20/20 Stella Koch Ocker 
    * initial conversion f77 --> python
01/23/20 -- now reads input parameters from dictionary program ne2001p_input
02/08/20 -- JMC
    * imports config_ne2001p for  model setup
    * rsun = 8.5 commented out; now set in setup file
02/09/20 -- JMC
    * comment out hitvoidflag; not used even in Fortran code
02/10/20 -- JMC
    * definition of void arrays moved to config_ne2001p.py
'''

from mwprop.nemod.config_nemod import *

def nevoidN(x,y,z):

    nevN = 0.
    FvN = 0.
    hitvoid = 0
    wvoid = 0

    '''
    note rotation matrix in the 'q = ' statement below
    corresponds to \Lambda_z\Lambda_y
    where \Lambda_y = rotation around y axis
        \Lambda_z = rotation around z axis
    defined as
        \Lambda_y =  c1  0  s1
                     0  1   0
                    -s1  0  c1

        \Lambda_z =  c2 s2   0
                    -s2 c2   0
                    0  0   1
        =>
        \Lambda_z\Lambda_y =  c1*c2   s2   s1*c2
                            -s2*c1   c2  -s1*s2
                            -s1    0      c1
    so the rotation is around the y axis first, then the z axis
    '''

    for j in range(nvoids):
        dx = x-xv[j]
        dy = y-yv[j]
        dz = z-zv[j]
        q = (cc12[j]*dx + s2[j]*dy + cs21[j]*dz)**2. / aav[j]**2. + (-cs12[j]*dx + c2[j]*dy - ss12[j]*dz)**2. / bbv[j]**2. + (-s1[j]*dx + c1[j]*dz)**2. / ccv[j]**2.
        if edgev[j] == 0. and q < 3.: # note this doesn't actually get used in NE2001; no clumps with edge = 0
            nevN = nev[j] * exp(-q)
            FvN = Fv[j]
            hitvoid = j+1
            #hitvoidflag[j] = 1
            #print('x,y,z,nev',x,y,z,nevN)
        if edgev[j] == 1. and q <= 1.:
            #print('void no.', j,'q', q)
            nevN = nev[j]
            FvN = Fv[j]
            hitvoid = j+1 # 3/6/22 -- SKO changed this from hitvoid = j --- j = 0 for Gum edge, which means it doesn't get counted
            #hitvoidflag[j] = 1
            #print('x,y,z,nev',x,y,z,nevN)

    
    if hitvoid != 0:
        wvoid = 1
    
    #print(hitvoid,wvoid)  
    return nevN, FvN, hitvoid, wvoid



# mwprop v2.0 Jan 2026

# Reads NE20x input data into dictionaries

from __future__ import print_function
import numpy as np
import os
import warnings

from scipy.interpolate import CubicSpline

loadtxt = np.loadtxt
genfromtxt = np.genfromtxt

warnings.filterwarnings('ignore','.*Input line 142.*')
warnings.filterwarnings('ignore','.*Input line 16.*')

def read_nemod_parameters():
    """
    Reads standard input files for the model
    and puts them into dictionaries

    NB. Entries like name = name.replace(',', '') take into account
    that the original NE2001 fortran input files have occasional ',' that
    need to be parsed out to get consistent dictionary keys

    JMC 2020 Jan 12-13
    """

    # Note absolute path

    inpdir = os.path.dirname(os.path.realpath(__file__))+'/params/'

    which_mod = loadtxt(inpdir+'which_model.inp',dtype=str)
    if which_mod=='NE2025':
    	eval_NE2025 = True
    	eval_NE2001 = False
    if which_mod=='NE2001':
    	eval_NE2001 = True
    	eval_NE2025 = False
    if eval_NE2001:
        galfile = inpdir + 'gal01.inp'
        negcfile = inpdir + 'ne_gc.NE2001.inp'
        clumpfile = inpdir + 'neclumpN.NE2001.dat'  
        voidfile = inpdir + 'nevoidN.NE2001.dat'
    if eval_NE2025:
        galfile = inpdir + 'gal25.inp'
        negcfile = inpdir + 'ne_gc.NE2025.inp'
        clumpfile = inpdir + 'neclumpN.NE2025.dat'  
        voidfile = inpdir + 'nevoidN.NE2025.dat'

    armsfile = inpdir + 'ne_arms_log_mod.inp'
    nelismfile = inpdir +'nelism.inp'

    # --------------------------------------------------------------------

    # Dictionary from gal01.inp

    Dgal = {}

    # weights:
    names = \
        loadtxt(galfile, skiprows=1, max_rows=1, \
                usecols=(7,8,9,10,11,12,13), dtype=str)
    values = \
        genfromtxt(galfile, skip_header=1, max_rows=1, \
                   usecols=(0,1,2,3,4,5,6), dtype=int)

    for n, name in enumerate(names):
        name = name.replace(',', '')
        Dgal[name] = values[n]


    # component parameters:
    names = loadtxt(galfile, skiprows=2, usecols=(0,), delimiter=':', dtype=str)
    values = loadtxt(galfile, skiprows=2, usecols=(1,), delimiter=':', dtype=float)

    for n, name in enumerate(names):
        Dgal[name] = values[n]

    # --------------------------------------------------------------------

    # Arrays from ne_arms_log_mod.inp 

    Darms = {}
    names = loadtxt(armsfile, skiprows=1, max_rows=1, usecols=(0,1,2,3), dtype=str)
    #aarm, rmin, thmin, extent = \
    xxx = loadtxt(armsfile, skiprows=2, unpack=True, \
                  usecols=(0,1,2,3), dtype=float)
    for n, name in enumerate(names):
        Darms[name] = xxx[n]  

    # --------------------------------------------------------------------

    # Dictionary for Galactic center component
    names1 = loadtxt(negcfile, skiprows=1, usecols=(3,4,5), max_rows=1, dtype=str)
    values1 = loadtxt(negcfile, delimiter=',', skiprows=1, usecols=(0,1,2), max_rows=1, dtype=float)

    names2 = loadtxt(negcfile, skiprows=2, usecols=(1,), dtype=str)
    values2 = loadtxt(negcfile, skiprows=2, usecols=(0,), dtype=float)

    names = np.concatenate((names1, names2))
    values = np.concatenate((values1, values2))

    Dgc = {}
    for n, name in enumerate(names):
        name = name.replace(',', '')
        if 'zgc' in name: name = 'zgc'   
        Dgc[name] = values[n]

    # --------------------------------------------------------------------
    
    # Local ISM
    # has different numbers of fields on each line, so use open + readline

    fism = open(nelismfile, 'r')
    fism.readline()         # Skip comment line
    line = 'x'

    Dlism = {}
    while line != '':
       line  = fism.readline()
       line = line.replace(',', '')
       line = line.replace('\n', '')
       x = line.split(' ')
       N = np.size(x)
       if N > 1:
           N2 = int(N/2)
           fields = x[N2:]
           values = np.array(x[0:N2]).astype(float)
           for n, name in enumerate(fields):
               Dlism[name] = values[n]
    fism.close()
    
    # --------------------------------------------------------------------

    # clumps:
    fields = loadtxt(clumpfile, unpack=True, usecols=(0,1,2,3,4,5,6), max_rows=1, dtype=str) 
    type, names = loadtxt(clumpfile, skiprows=1, usecols=(8, 9), unpack=True,  dtype=str)
    values = loadtxt(clumpfile, skiprows=1, unpack=True, usecols=(1,2,3,4,5,6,7), dtype=float)

    Dclumps = {}
    for n, name in enumerate(names):
        Dclumps[name] = {}
        for m, field in enumerate(fields):
             Dclumps[name][field] = values[m, n]
        
    # --------------------------------------------------------------------

    # voids:
    fields = loadtxt(voidfile, unpack=True, usecols=(0,1,2,3,4,5,6,7,8,9,10), max_rows=1, dtype=str) 
    names = loadtxt(voidfile, skiprows=1, usecols=(12,), unpack=True,  dtype=str)
    values = loadtxt(voidfile, skiprows=1, unpack=True, usecols=(1,2,3,4,5,6,7,8,9,10,11), dtype=float)

    Dvoids= {}
    for n, name in enumerate(names):
        Dvoids[name] = {}
        for m, field in enumerate(fields):
             Dvoids[name][field] = values[m, n]

    # --------------------------------------------------------------------

    return Dgal, Dgc, Dlism, Dclumps, Dvoids, Darms, eval_NE2025, eval_NE2001

if __name__ == '__main__':

    Dgal, Dgc, Dlism, Dclumps, Dvoids, Darms, eval_NE2025, eval_NE2001 = read_nemod_parameters()





# mwprop v2.0 Jan 2026

'''
clump subroutine -- put in density subfolder

comments copied from Fortran code:

returns electron density necN and fluctuation parameter FcN at position designated by l,b,d,x,y,z,c for a set of clumps with parameters read in from fild neclumpN.dat

input: x,y,z coordinates (kpc) (as in TC93)

output:
    necN            electron density in clump at (x,y,z)
    FcN             fluctuation parameter
    hitclump = 0:   no clump hit
            j>0:    jth clump hit

parameters:
    lc    = galactic longitude of clump center
    bc    = galactic latitude of clump center
    (xc,yc,zc) = clump center location (calculated)
    nec   = internal peak electron density
    rc    = clump radius at 1/e
    Fc    = clump fluctuation parameter
    edgec = 0 => use exponential rolloff out to 5rc
            1 => uniform and truncated at 1/e

lc,bc = Galactic coordinates (deg)
nec = clump electron density (cm^{-3})
Fc = fluctuation parameter
dc = clump distance from Earth (kpc)
rc = clump radius (kpc)
edgec = 0,1  0=> Gaussian, 1=> Gaussian w/ hard edge at e^{-1}
type = LOS type (P pulsar, G other Galactic, X extragalactic
losname = useful name

Version history:

01/18/2020 Stella Koch Ocker, initial conversion f77 --> python

01/23/20 -- JMC
    * now reads input parameters from dictionary program ne2001p_input

02/08/20 -- JMC
    * imports config_ne2001p for  model setup  
    * rsun = 8.5 now commented out, included in setup file

12/28/21 - 01/02/22 -- JMC
    * added relevant_clumps function to prefilter clump list for the line
      of sight before integrating along it

11/27/2023 -- SKO 
    * changed default value of rcmult to make sure clump prefiltering working properly
12/15/2023 -- SKO
    * corrected definition of inds_relevant so that LOS with dmax inside clump get counted 
'''

from mwprop.nemod.config_nemod import *

def relevant_clumps(l, b, dmax, rcmult): 
    """
    Identifies clumps that contribute to n_e for the LoS
    specified by ldeg, bdeg, dmax.

    Input:
        l, b = Galactic coordinates of LoS (rad)
        dmax = maximum distance to calculate for LoS (kpc)
        rcmult = multiplier of clump radius for proximity decision

        Clump parameters in scope:
        lc, bc = Galactic coordinates of clumps (deg)
        xc, yc, zc = coordinates of clump (kpc)
        slc, clc, sbc, cbc = sin and cos of lc, bc
        rc = radial scales of clumps (kpc)
        dc = clump distances (kpc)
    """
    # calculate 'closest' = closest approach of LoS to clump:
    cl = cos(l)
    sl = sin(l)
    cb = cos(b)
    sb = sin(b)
    cos_theta = cb*cbc*(sl*slc+cl*clc) + sb*sbc
    # SKO -- need to force inds_relevant to recognize LOS that pass straight through clump center
    straight = np.where((1-cos_theta)<1e-4)
    cos_theta[straight] = 0.9998 # SKO -- to avoid sin inf errors
    sin_theta = sqrt(1-cos_theta**2) 
    closest = dc * sin_theta
    closest[straight] = 0. # if you pass a LOS straight through the center, you want closest to be 0

    # want cos_theta > 0 because we want solutions where sin_theta is smaller than pi/2;
    # this will fail if there are ever very large clumps in the model.

    # note dc + rcmult*rc will be larger than dmax if dmax = dc
    # SKO 12-15-23: changed condition to dc - rcmult*rc < dmax 

    #inds_relevant = \
    #   np.where((dc + rcmult*rc < dmax) & (cos_theta > 0) & (closest < rcmult*rc)) # old condition

    inds_relevant = \
        np.where((dc-rcmult*rc < dmax) & (cos_theta > 0) & (closest < rcmult*rc))

    #print('dc',dc[inds_relevant[0]])
    #print('dmax', dmax)
    #print('rcmult',rcmult[inds_relevant[0]])

    if np.any(inds_relevant)==False: # if no relevant clumps
        inds_relevant = np.array([-1])

    return inds_relevant

def neclumpN(x,y,z, inds_relevant=None):
    """
    Returns electron density and F parameter for position x,y,z contributed by 
    any clumps along the LoS.   

    inds_relevant: 
        None:     step through all clumps in list initiated in config_ne2001p 
        Not None: function assumes this is a tuple indicating indices for
                  only those clumps relevant for the line of sight. 
    """

    if inds_relevant is None:
        clumpnums = range(0, nclumps) 
    else:
        clumpnums = inds_relevant[0]
    
    necN = 0.
    hitclump = 0
    FcN = 0.
    arg = 0.
    hitclumpflag = np.zeros(nclumps)

    #print('number of relevant clumps:',clumpnums)

    if np.isscalar(clumpnums)==True and clumpnums==-1: # SKO 11/27/23 -- if no relevant clumps, exit with necN set to 0
        return necN,FcN,hitclump,arg 

    else:
        for j in clumpnums: 
            arg = ((x-xc[j])**2. + (y-yc[j])**2. + (z-zc[j])**2.) / (rc[j]**2.)
            if edgec[j] == 0. and arg < 5.:
                necN = necN + nec[j] * exp(-arg)
                FcN = Fc[j]
                hitclump = j
                hitclumpflag[j] = 1
            if edgec[j] == 1. and arg <= 1.:
                #print('arg: ',arg,' edgec: ',edgec[j])
                #print('xc,yc,zc,rc',xc,yc,zc,rc) 
                necN = necN + nec[j]
                FcN = Fc[j]
                hitclump = j
                hitclumpflag[j] = 1

            
        return necN, FcN, hitclump, arg # SKO 3/6/22 -- added arg for debugging



# mwprop v2.0 Jan 2026

"""
NE2001 master script

Usage: 
    command line:  python NE2001.py l b DM_D ndir [-h]
    in ipython:    run NE2001.py l b DM_D ndir [-h]
    in script:     Dv, Du, Dd = ne2001(l, b, DM_D, ndir)


Change Log:

01/16/2020 initial code sent from SKO --> JMC
01/29/2020 JMC -- mods to make compatible with other new Python modules 
02/08/2020 JMC -- add import of ne2001p_config  
02/11/2020 JMC -- turn into function + main()
17/12/2020 JMC -- added plotting Cn^2 vs d 
01/05/2022 JMC -- moved some definitions to ne2001p_config file, redefined nu -> rf_nu 
01/07/2022 JMC -- restructured to use new dmdsm routines,
                  ddmdsm_dm2d_ne2001.py and mdsm_d2dm_ne2001.py
01/09/2022 JMC -- restructed to take plotting outside of the ne2001 function
                  merged dmdsm_dm2d_ne2001.py + dmdsm_d2dm_ne2001.py -> dmdsm_ne2001p.py
03/01/2022 SKO -- merged into package build
03/2022 - 05/2022 SKO -- clump debugging
11/27/2023 SKO -- corrected clump filtering, modified rcmult
12/14/2023 SKO -- all output files (plots, text files) now save to 'output_ne2001p' directory in user's current working directory
01/09/2026 SKO -- NE2001.py and NE2025.py now separate master scripts
"""

import sys,os

global eval_NE2025,eval_NE2001
eval_NE2025 = False
eval_NE2001 = True

# set which_model.inp
inpdir = os.path.dirname(os.path.realpath(__file__))+'/params/'
with open(inpdir+"which_model.inp", "w") as f:
    f.write("NE2001")
f.close()
#with open(inpdir+"which_model.inp", "r") as f:
#    print(f.read())
#f.close()


# ------------------------------------------------------------
# config_nemod sets up the model with all dictionaries etc: 
# ------------------------------------------------------------
from mwprop.nemod.config_nemod import *

# ----------------------------------------
# Functions to do line of sight integrals:
# ----------------------------------------
from mwprop.nemod.dmdsm import *

# ------------------------------------------------------
# Functions to calculated derived scattering quantities:
# ------------------------------------------------------
from mwprop.scattering_functions import scattering_functions2020 as sf

# ---------------------------------------------------------------------
# for additional FRB scattering calculations (integrated to 'infinity')
# ---------------------------------------------------------------------
from mwprop.scattering_functions import iss_mw_utils2020p_mod as iss_utils


"""
Returns values in two ways:
    * prints to std out in same format as NE2001/Fortran (with some additional lines)
    * dictionaries Dn, Dv, Du, Dd containing names, values, units, and descriptions

To add:
Options:
- Classic option: all output in regular format
- DM->D or D->DM only: no scattering output
- do_analysis option: diagnostic LoS info (slows down the calculation); written to f24, f25 files
- extragalactic LoS: 
    * A specified D > 300 kpc (say) implies extragalactic (beyond MW halo)
    * option to include halo estimate of DM
Help file: print out assumptions and other info about the code (README file?)
"""
script_path = os.path.dirname(os.path.realpath(__file__))

basename = sys.argv[0].split('/')[-1].split('.')[0]
now = datetime.datetime.now()
plotstamp = basename + '_' + str(now).split('.')[0]

global rsun, rf_ref, vperp, sikol, louter, linner, ds_fine, ds_coarse
global Dgal01, Dgc, Dlism, Dclumps, Dvoids, Darms, Darmmap, armmap, \
           r1, th1, th1deg, coarse_arms, rarray, tharray, \
           armsplines, armarray, tangents, normals, curvatures

global wg1, wg2, wga, wggc, wglism, wgcN, wgvN
global n1h1, h1, A1, F1, n2, h2, A2, F2, na, ha, wa, Aa, Fa
global apldr, bpldr, cpldr, dpldr, yldrmin, yldrmax
global aplsb, bplsb, cplsb, dplsb, ylsbmin, lsbmax
global alhb, blhb, clhb, xlhb, ylhb, zlhb, thetalhb, nelhb0, Flhb, ylhbmin, ylhbmax
global xlpI, ylpI, zlpI, rlpI, drlpI, nelpI, dnelpI, FlpI, dFlpI
global y_lism_min, y_lism_max
global xgc, ygc, zpgc, rgc, hgc, negc0, Fgc0


# --------------------------------------------------------------------

def ne2001(ldeg, bdeg, dmd, ndir, 
    classic=True, dmd_only=True, 
    do_analysis=False, plotting=False, verbose=False, debug=False,eval_NE2001=True,eval_NE2025=False):
    """
    Calculates d from DM or DM from d for ndir >= 0 or ndir < 0

    Input:

    ldeg, bdeg = Galactic longitude and latitude in deg
    dmd = DM (pc/cc) or d (kpc)
    ndir >= 0   DM -> d
          < 0    d -> DM
    dmd_only = True     => compute only DM or distance, no scattering quantities
    classic = True      => print output in format as in Fortran version
    do_analysis = True  => calculate and output diagnostic quantities for the line of sight
    plotting = True     => plot DM vs distance along LoS, SM also if dmd_only = False     
    verbose = True      => print results to std out 
                           (note not same as 'verbose' internal to functions 

    Output: two types:
        Classic: All input and output quantities for DM/D, scattering, scintillations, 
                                            emission measure, ...
        Dictionaries for output quantities.
    """

    if do_analysis or plotting: # make directory for outputs (need this block here for script calls to ne2001())
        try:
            os.mkdir(os.getcwd()+'/output_ne2001p')
        except FileExistsError:
            pass

    num_outputs = where(dmd_only, 3, 17)

    # Print input parameters before executing code

    if classic:
        print('#NE2001 input: 4 parameters')
        print('{:<1s}{:>10.4f}{:>13s}{:>10s}{:<25s}{:35s}'.format(
            '', ldeg, 'l   ', '', '(deg)', 'GalacticLongitude'))
        print('{:<1s}{:>10.4f}{:>13s}{:>10s}{:<25s}{:35s}'.format(
            '', bdeg, 'b   ', '', '(deg)', 'GalacticLatitude'))
        print('{:<1s}{:>10.4f}{:>13s}{:>10s}{:<25s}{:35s}'.format(
            '', dmd,  'DM/D', '', '(pc cm^{-3}_or_kpc)', 'Input DM or Distance'))
        print('{:<1s}{:>10d}{:>13s}{:>10s}{:<25s}{:35s}'.format(
            '',ndir, 'ndir', '', '1:DM->D;-1:D->DM', 'Which?(DM or D)'))
        print('#NE2001 output: %d values'%(num_outputs))

    # ---------------------------
    # Calculate distance from DM:
    # ---------------------------
    if ndir >= 0:
        dm_target = dmd
        #limit, distmodel, dmpsr, sm, smtau, smtheta, smiso \
        model_out =  dmdsm_dm2d(deg2rad(ldeg), deg2rad(bdeg), dm_target, 
                     ds_coarse=ds_coarse, ds_fine=ds_fine, Nsmin=10, 
                     dm2d_only=dmd_only, do_analysis=do_analysis, 
                     plotting=plotting, verbose=verbose, debug=debug)
        limit_d = model_out[0]
        d_mod = model_out[1]
        dm_mod = model_out[2]

        if limit_d == ' ':
            dmlabel = 'DispersionMeasure'
        else:
            dmlabel = 'DispersionMeasureAsymptote'
        dlabel = 'ModelDistance'

        if dmd_only is False:           # => scattering parameters are calculated
            sm = model_out[3] 
            smtau = model_out[4] 
            smtheta = model_out[5] 
            smiso = model_out[6] 

    # ---------------------------
    # Calculate DM from distance:
    # ---------------------------
    if ndir <  0:
        d_target = dmd
        model_out = dmdsm_d2dm(deg2rad(ldeg), deg2rad(bdeg), d_target,
                    ds_coarse=ds_coarse, ds_fine=ds_fine, Nsmin=10,
                    d2dm_only=dmd_only, do_analysis=do_analysis, 
                    plotting=plotting, verbose=verbose)

        # The first three entries in model_out are limit, d, and dm for all cases
        limit_d = model_out[0]
        d_mod = model_out[1]
        dm_mod = model_out[2]

        if dmd_only is False:           # => scattering parameters are calculated
            sm = model_out[3] 
            smtau = model_out[4] 
            smtheta = model_out[5] 
            smiso = model_out[6] 


        dmlabel = 'ModelDM'
        dlabel = 'Distance'

    dmz = dm_mod*abs(sin(deg2rad(bdeg)))

    if classic:
        print('{:<1s}{:>10.4f}{:>13s}{:>9s} {:<25s}{:35s}'.format(
            limit_d, d_mod,'DIST', '', '(kpc)', dlabel))
        print('{:<1s}{:>10.4f}{:>13s}{:>9s} {:<25s}{:35s}'.format(
            '',dm_mod,'DM  ', '', '(pc cm^{-3})',dmlabel))
        print('{:<1s}{:>10.4f}{:>13s}{:>9s} {:<25s}{:35s}'.format(
            '',dmz,'DMz ', '', '(pc cm^{-3})','DM_Zcomponent'))

    # Also print out scattering, scintillation etc. parameters if requested
    if dmd_only is False:
        tau = sf.tauiss(d_mod, smtau, rf_ref)               
        sbw = sf.scintbw(d_mod, smtau, rf_ref)
        stime = sf.scintime(smtau, rf_ref, vperp)
        theta_x = sf.theta_xgal(sm, rf_ref)
        theta_g = sf.theta_gal(smtheta, rf_ref)
        transfreq = sf.transition_frequency(sm, smtau, smtheta, d_mod)
        emsm = sf.em(sm, louter=louter, si=sikol)
        tau_x,sbw_x,deffsm2 = sf.taux_from_thetax(d_mod,sm,smtau,smtheta,theta_x,tau,sbw)

        if classic:
            print('{:<1s}{:>8.4e}{:>17s}{:>6s}{:<25s}{:35s}'.format(
                '',sm,'SM      ', '', '(kpc m^{-20/3})','ScatteringMeasure'))
            print('{:<1s}{:>8.4e}{:>17s}{:>6s}{:<25s}{:35s}'.format(
                '',smtau,'SMtau   ', '','(kpc m^{-20/3})','SM_PulseBroadening'))
            print('{:<1s}{:>8.4e}{:>17s}{:>6s}{:25s}{:35s}'.format(
                '',smtheta,'SMtheta ',  '','(kpc m^{-20/3})','SM_GalAngularBroadening'))
            print('{:<1s}{:>8.4e}{:>17s}{:>6s}{:25s}{:35s}'.format(
                '',smiso,'SMiso   ', '','(kpc m^{-20/3})','SM_IsoplanaticAngle'))
            print('{:<1s}{:>8.4e}{:>17s}{:>6s}{:25s}{:35s}'.format(
                '',emsm,'EM      ', '','(pc cm^{-6})','EmissionMeasure_from_SM'))
            print('{:<1s}{:>8.4e}{:>17s}{:>6s}{:25s}{:35s}'.format(
                '',tau,'TAU     ', '','(ms)','PulseBroadening @1GHz'))
            print('{:<1s}{:>8.4e}{:>17s}{:>6s}{:25s}{:35s}'.format(
                '',sbw, 'SBW     ',  '','(MHz)','ScintBW @1GHz'))
            print('{:<1s}{:>8.4e}{:>17s}{:>6s}{:25s}{:35s}'.format(
                '',stime,'SCINTIME',  '','(s)','ScintTime @1GHz @100 km/s'))
            print('{:<1s}{:>8.4e}{:>17s}{:>6s}{:25s}{:35s}'.format(
                '',theta_g,'THETA_G ', '', '(mas)','AngBroadeningGal @1GHz'))
            print('{:<1s}{:>8.4e}{:>17s}{:>6s}{:25s}{:35s}'.format(
                '',theta_x,'THETA_X ', '', '(mas)','AngBroadeningXgal @1GHz'))
            print('{:<1s}{:>10.4f}{:>17s}{:>6s}{:25s}{:35s}'.format(
                '',transfreq,'NU_T    ', '', '(GHz)','TransitionFrequency'))
            print('{:<1s}{:>10.4f}{:>17s}{:>6s}{:<25s}{:35s}'.format(
                '',deffsm2,'DEFFSM2 ', '','(kpc)','EffectiveScreenDistance'))
            print('{:<1s}{:>8.4e}{:>17s}{:>6s}{:<25s}{:35s}'.format(
                '',tau_x,'TAU_X   ', '','(ms)','XGalPulseBroadening@1GHz'))
            print('{:<1s}{:>8.4e}{:>17s}{:>6s}{:<25s}{:35s}'.format(
                '',sbw_x, 'SBW_X   ',  '','(MHz)','XGalScintBW@1GHz'))

    # Dictionaries of variable names, values, units, and descriptions
    Dn, Dv, Du, Dd = {}, {}, {}, {}

    if dmd_only is True:
        # DM and distance only:
        Dkeyvalues = list([ldeg, bdeg, dmd, ndir, limit_d, d_mod , dm_mod, dmz])

        Dkeynames = np.array(['l', 'b', 'DM/D', 'ndir', 'limdist', 'DIST', 'DM', 'DMz'])

        Dkeyunits = np.array(['deg', 'deg', 'pc-cm^{-3}_or_kpc', 
            '1:DM->D;-1:D->DM', 'blank_or_>>', 'kpc', 'pc-cm^{-3}', 'pc-cm^{-3}'])

        Dkeydesc = np.array(['GalacticLongitude', 'GalacticLatitude', 'Input_DM_or_Distance',
           'Which?(DM_or_D)', 'LowerDistanceLimit', 'ModelDistance', 'DispersionMeasure',
           'DM_Zcomponent'])

        for n, key in enumerate(Dkeynames):
           Dn[key] = key
           Dv[key] = Dkeyvalues[n]
           Du[key] = Dkeyunits[n]
           Dd[key] = Dkeydesc[n]

    else:
    # DM, distance, and scattering output:

        Dkeyvalues = list([ldeg, bdeg, dmd, ndir, limit_d, d_mod, dm_mod, dmz, 
            sm, smtau, smtheta, smiso, tau, sbw, stime, theta_g, theta_x, 
            transfreq, emsm, deffsm2, tau_x, sbw_x])

        Dkeynames = np.array(['l', 'b', 'DM/D', 'ndir', 'limdist', 'DIST', 'DM', 'DMz', 
            'SM', 'SMtau', 'SMtheta', 'SMiso', 'TAU', 'SBW', 'SCINTIME', 'THETA_G', 'THETA_X', 
            'NU_T', 'EM', 'DEFFSM2', 'TAU_X', 'SBW_X'])

        Dkeyunits = np.array(['deg', 'deg', 'pc-cm^{-3}_or_kpc', '1:DM->D;-1:D->DM', 
            'blank_or_>>', 'kpc', 'pc-cm^{-3}', 'pc-cm^{-3}', 'kpc-m^{-20/3}', 'kpc-m^{-20/3}',
           'kpc-m^{-20/3}', 'kpc-m^{-20/3}',  'ms', 'MHz', 's',
           'mas', 'mas', 'GHz', 'pc-cm^{-6}', 'kpc', 'ms', 'MHz'])

        Dkeydesc = np.array(['GalacticLongitude', 'GalacticLatitude', 'Input_DM_or_Distance',
           'Which?(DM_or_D)', 'LowerDistanceLimit', 'ModelDistance', 'DispersionMeasure',
           'DM_Zcomponent', 'ScatteringMeasure', 'SM_PulseBroadening',
           'SM_GalAngularBroadening', 'SM_IsoplanaticAngle',
           'PulseBroadening@1GHz', 'ScintBW@1GHz',
           'ScintTime@1GHz@100km/s', 'AngBroadeningGal@1GHz', 'AngBroadeningXgal@1GHz',
           'TransitionFrequency', 'EmissionMeasure_from_SM@outer1pc', 
           'EffectiveScreenDistance', 'XGalPulseBroadening@1GHz', 'XGalScintBW@1GHz'])

        for n, key in enumerate(Dkeynames):
           Dn[key] = key
           Dv[key] = Dkeyvalues[n]
           Du[key] = Dkeyunits[n]
           Dd[key] = Dkeydesc[n]

    return Dn, Dv, Du, Dd

# --------------------------------------------------------------------

def plot_dm_ne_cn2_vs_d(Dv, f25file = 'f25_dm2d_ne_dsm_vs_s.txt'):
    # Plot if requested and calc_scattering = True:
       s, ne, dsm, dmvss = np.loadtxt(f25file, unpack=True, usecols = (0, 4, 5, 10), skiprows=3)
       DMmax = dmvss.max()

       # For plot labels:
       dm = Dv['DM']                    
       deffsm2 = Dv['DEFFSM2']
       ldeg = Dv['l']
       bdeg = Dv['b']
           
       indmax = np.where(ne != 0)[0][-1]
       indkeep = min(int(1.1*indmax), np.size(ne))
       s = s[0:indkeep]
       ne = ne[0:indkeep]
       cn2  = sm_factor *  dsm[0:indkeep]
       dmvss = dmvss[0:indkeep]

       dbar_ne = np.trapz(s*ne) / np.trapz(ne)
       dbar_ne2 = np.trapz(s*ne**2) / np.trapz(ne**2)

       fig = figure(figsize=(6,6))
       subplots_adjust(left=0.15, bottom=0.15)

       ax = fig.add_subplot(311)
       plot(s, dmvss)
       ylabel(r'$\rm DM \ (pc\ cm^{-3}) $', fontsize=13)
       annotate(r'$l = %6.1f^{\circ}  \ \ b = %5.1f$'%(ldeg, bdeg), 
           xy=(0.025, 0.875), xycoords='axes fraction', ha='left', va='center', fontsize=10)
       title(r'$\overline{d}_{n_e} = %8.2f \ \ \ \ \ \overline{d}_{n_e^2} = %8.2f \ \ \ \ \ \overline{d}_{\rm SM} = %8.2f \ \ \ \ \ {\rm DM_{max}} =  %8.2f$'%(dbar_ne, dbar_ne2, deffsm2, DMmax))
       tick_params(labelbottom = False)

       ax = fig.add_subplot(312)
       plot(s, ne, label=r'$\rm n_e$')
       ylabel(r'$n_e \ (\rm cm^{-3}) $', fontsize=13)
       tick_params(labelbottom = False)

       ax = fig.add_subplot(313)
       plot(s, cn2, label=r'$\rm C_n^2$')
       ylabel(r'$C_{\rm n}^2 \ (\rm m^{-20/3}) $', fontsize=13)
       xlabel(r'$\rm Distance \ along \ LoS \ \ (kpc) $', fontsize=13)
       annotate(plotstamp, xy=(0.70, 0.02), xycoords='figure fraction', ha='left', va='center', fontsize=5)
       fig.tight_layout()

       output_dir = os.getcwd()+'/output_ne2001p/'

       plotfile = \
            'DM_ne_Cn2_vs_d_l_b_dm_' + '_' + basename + '.pdf'
       savefig(output_dir+plotfile)
       show()

       return

# --------------------------------------------------------------------

# Main

if __name__ == '__main__':

    # If 'explain' option is set, print out README file and exit.
    if '-e' in sys.argv or '--explain' in sys.argv:
         script_path = os.path.dirname(os.path.realpath(__file__)) # SKO -- so that README will always be found
         infile = script_path+'/README.txt' # removed ne2001p 
         with open(infile) as fexplain:
             print(fexplain.read())
             fexplain.close()
         #print(open(infile, 'r').read())
         sys.exit()

    try:
        # parse command line arguments and execute
        parser = argparse.ArgumentParser(description='NE2001 (python version 01/05/22)')
        parser.add_argument('l',help='Galactic longitude (deg)')
        parser.add_argument('b',help='Galactic latitude (deg)')
        parser.add_argument('DM_D',help='DM (pc cm^{-3}) or distance (kpc)')
        parser.add_argument('ndir',help='ndir = 1 (DM->D), ndir = -1 (D->DM)')

        parser.add_argument('-a', '--analysis', action='store_true', 
            help='Do line of sight analysis')

        parser.add_argument('-p', '--plotting', action='store_true', 
            help='Do diagnostic plotting')

        parser.add_argument('-s', '--scattering', action='store_true', 
            help='Calculate scattering and scintillation parameters')

        parser.add_argument('-m', '--modern', action='store_true', 
            help='Modern output (turns off classic output like Fortran version)')

        parser.add_argument('-v', '--verbose', action='store_true', 
            help='Verbose output (classic output of Fortran version)')

        parser.add_argument('-b', '--debug', action='store_true', 
            help='debug output')
            
        args = parser.parse_args()


    except SystemExit:
        print('Try again with inputs')
        print('Use NE2001 -e to get explanation of code')
        sys.exit()

    ldeg = float(args.l)
    bdeg = float(args.b)
    dmd = float(args.DM_D)
    ndir = int(args.ndir)
    do_analysis = args.analysis
    do_plotting = args.plotting
    calc_scattering = args.scattering
    dmd_only = not calc_scattering
    verbose = args.verbose
    debug = args.debug
    modern = args.modern
    eval_NE2001 = True
    eval_NE2025 = False
    
    if modern:
        classic = False
    else:
        classic = True

    Dn, Dv, Du, Dd = ne2001(ldeg, bdeg, dmd, ndir, 
        classic=classic, verbose=verbose, dmd_only=dmd_only, 
        do_analysis=True, plotting=do_plotting, debug=debug,eval_NE2001=eval_NE2001,eval_NE2025=eval_NE2025)

    if calc_scattering and do_plotting:
        output_dir = os.getcwd()+'/output_ne2001p/'
        plot_dm_ne_cn2_vs_d(Dv, f25file = output_dir+'f25_dm2d_ne_dsm_vs_s.txt')

        input('hit return')
        close('all')




# mwprop v2.0 Jan 2026

"""
Python versions of functions in scattering98.f that is
part of the NE2001 Fortran package

2019 December 31

Added: 
    transition_frequency_from_obs
    sm_from_tau

2026 Feb 2: Moved tau_x calculation here
------
Notes in 1998-2001 Fortran version: 
This version from 18 March 1998 uses revised coefficients 
that are consistent with Cordes \& Rickett (1998, ApJ)

Note that scaling laws are explicitly for a Kolmogorov medium 
in the strong but not superstrong regime 
(as defined in Cordes and Lazio 1991)

Removed theta_iso_test (was redundant)

Modifications:
28 March 2001: added FUNCTION TRANSITION_FREQUENCY 
"""

from __future__ import print_function

from math import pi

from mwprop.get_constants_waveprop_astropy_version import *

import numpy as np

def fbeta(si=11./3.):
    """
    Calculates the f(\beta) function used in scattering calculations
    for a power-law wavenumber spectrum with spectral index si == beta.
    [nb. beta not used here to avoid confusion with the beta PDF]

    Also note that other 'fbeta' functions in the literature
    are the same as here but multiplied by 4*pi**2

    Cordes & Lazio 2003  NE2001 Paper II, Appendix Eq A5
    JMC 2020 Jan 1
    """ 
    from math import gamma

    beta2 = si / 2.
    beta1 = si - 1
    fbeta = (gamma(2-beta2) * gamma(beta2-1)) / (gamma(beta2)**2 * 2**beta1) 
    return fbeta

def k_eta(si=11./3.):
    """
    Calculates the K_\eta(\beta) function used in scattering calculations
    for a power-law wavenumber spectrum with spectral index si == beta.
    [nb. beta not used here to avoid confusion with the beta PDF]

    From A-O book snippet (Drafted in 2019)
    'Estimating the Wavenumber Spectrum for Electron Density in the ISM'
    Eq 29.

    JMC 2020 Jan 1
    """ 
    from math import gamma

    beta2 = si / 2.
    beta4 = 4 - si

    keta = (2.*pi)**beta4 * gamma(3-beta2) / beta4

    return keta
    

def tauiss(d, sm, nu):

    """
    calculates the pulse broadening time in ms
    from distance, scattering measure, and radio frequency

    input:      d = pulsar distance       (kpc)    
               sm = scattering measure    (kpm^{-20/3})
               nu = radio frequency       (GHz)
    output: tauss = pulse broadening time (ms) 
    """

    tauiss = 1000. * (sm / 292.)**1.2 * d * nu**(-4.4)
    return tauiss
    
def scintbw(d, sm, nu, C1 = 1.16):

    """
    calculates the scintillation bandwidth in kHz 
    from distance, scattering measure, and radio frequency

    input:        d = pulsar distance       (kpc)    
                 sm = scattering measure    (kpc m^{-20/3})
                 nu = radio frequency       (GHz)
                   C1 = 1.16 = dimensionless constant in sbw-tau relation 
    output: scintbw = scintillation bandwidth (MHz) (cf scattering98: kHz)
    """

    tau = tauiss(d, sm, nu)                       # ms
    scintbw = 1.e-3 * C1 / (2. * pi * tau)
    return scintbw
 
def scintime(sm, nu, vperp=100):
    """
     
    calculates the scintillation speed for given distance, galactic
    longitude and latitude, frequency, and transverse velocity      
     
    input:   sm = scattering measure    (kpc m^{-20/3})
             nu = radio frequency   (GHz)
          vperp = psr transverse speed      (km/s)  
                  (default 100)
     
    output: scintime = scintillation time (sec)
    
    usage: should be called with sm = smtau for appropriate
           line of sight weighting
    reference: eqn (46) of Cordes & Lazio 1991, ApJ, 376, 123.
    """

    scintime = 3.3 * nu**1.2 * sm**(-0.6) * (100./vperp)
    return scintime
 
 
def specbroad(sm, nu, vperp):
    """
     
    calculates the bandwdith of spectral broadening
    for given scattering measure, , frequency, and transverse velocity      
     
    input:   sm = scattering measure    (kpc m^{-20/3})
             nu = radio frequency   (GHz)
          vperp = psr transverse speed      (km/s)  
     
    output: specbroad = spectral broadening bandwidth (Hz)
    
    usage: should be called with sm = smtau for appropriate
           line of sight weighting
    reference: eqn (47) of Cordes & Lazio 1991, ApJ, 376, 123.
     
    nb: 
    The coeff. in the following line was 0.14 Hz from Cordes & Lazio (1991)
    It is changed to 0.097 to conform with FUNCTION SCINTIME and
    a new calculation consistent with Cordes & Rickett (1998)
    """

    specbroad = 0.097 * nu**(-1.2) * sm**0.6 * (vperp/100.) # Hz
    return specbroad
 
 
def theta_xgal(sm, nu):
    """
     
    calculates angular broadening for an extragalactic
    source of plane waves
     
    sm = scattering measure
    nu = radio frequency
    theta_xgal = angular broadening FWHM (mas)
     
    """
    theta_xgal = 128. * sm**0.6 * nu**(-2.2)
    return theta_xgal

def theta_gal(sm, nu):
    """
    calculates angular broadening for a galactic
    source of spherical waves
     
    sm = scattering measure
    nu = radio frequency
    theta_gal = angular broadening FWHM (mas)
    """

    theta_gal = 71. * sm**0.6 * nu**(-2.2)
    return theta_gal
 
def em(sm, louter=1., si=11./3.):

    """
    calculates the emission measure from the scattering measure
    using an assumed outer scale and spectral index of the
    wavenumber spectrum.

    Input:
        sm = scattering measure     (kpm^{-20/3})
        louter = outer scale        (pc)
        si = wavenumber spectral index (11/3 for Kolmogorov spectrum; 
                                        the only valid value for now)
    Output:
        em = emission measure       (pcm^{-6})
    
    For a wavenumber spectrum P_n(q) = q^{-alpha} from q_0 to q_1
    the mean square electron density is
     
    <n_e^2> =~  4pi*[C_n^2 / (alpha - 3) ] * q_0^{3 - alpha)
     
    (an approximate form that assumes (q_0 / q_1)^{3-alpha} >> 1.
     
    Jim Cordes 18 Dec 1989
    """
    em = sm \
       * ( (4. * pi * 1000.) / (si-3.) ) \
       * (louter*pc/ (2. * pi) )**(si-3.) \
       * (0.01) ** (20./3.)
    return em

def theta_iso(smiso, nu):
    """
    Input:
       smiso in (kpm^{-20/3}) x kpc^{5/3}
       nu in GHz
    Output:
       isoplanatiangle in microarcsec

    Requires:
       re = classical electron radius (cm)
       kpin cm
    12 October 1998 
    JMC

    \theta_{iso} = \delta r_s / d
                 = \left[
                         (\lambda r_e)^2 f_{\alpha} SM_{iso}
                  \right]^{1/\alpha}
    where \alpha = 5/3 for Kolmogorov case.
    NB SM_{iso} = \int_0^d ds s^{\alpha} \cnsq
       so SM_{iso} does not have the units of scattering
       measure, but rather units of SM x Length^{\alpha}
     
    f_{\alpha} = 8\pi^2 \Gamma(1-\alpha/2) / [\alpha 2^{\alpha} \Gamma(1+\alpha/2)]
    for \alpha = 5/3, f_{\alpha}= 88.3

    Constants in equation: 
          2*0.6*log10(30cm*r_e  = 13.287
          0.6*log10(f_alpha)   = 1.1676
          1.6 * alog10(kpc)    = 34.383
          -(20/3)*log(100)     = 8 ??? CHECK (2020)
    """

    theta_log_radian = \
          13.287 \
        + 1.2 * alog10(nu) \
        - 1.1676 \
        - 0.6 * alog10(smiso) \
        - 34.383 \
        + 8.                    

    # log10(microarsec/rad) = 11.314425
    theta_log_microarcsec = theta_log_radian + 11.314425    
    theta_iso = 10.**theta_log_microarcsec
    return theta_iso

def transition_frequency(sm, smtau, smtheta, dintegrate):
    """
    Returns the transition frequency between weak and strong scattering
    28 March 2001
    JMC

    input:
    (all sm values in (kpc m^{-20/3}))
              sm = int[\cnsq]
              smtau = int[(s/D)(1-s/D) \cnsq]
              smtheta = int[ (1-s/D) \cnsq]
              dintegrate = distance used to integrate \cnsq (kpc)
    output:
      transition_frequency = GHz given by
        \nu_t  = 318 GHz \\xi^{10/17}  SM^{6/17} D_{eff}^{5/17}  
        (nb. double \\ needed on xi to avoid a unicode error in python) 
      where
           D_{eff} = effective path length through medium
           D_{eff} = \int_0^dintegrate ds s \cnsq / \int_0^dintegrate ds  \cnsq

      Note we can calculate D_{eff} using
           D_{eff} = dintegrate * (sm - smtau/6 - smtau/3) / sm
    """

    xi= 0.3989                  # (2.*pi)^{-1/2} = fresnel scale definition factor
    coefficient=318.            # GHz; see NE2001 paper
    deff = (dintegrate*(sm - smtau/6. - smtheta/3.)) / sm 
    transition_frequency \
         = coefficient * xi**(10./17.) * sm**(6./17.) * deff**(5./17.) 
    return transition_frequency

def transition_frequency_from_obs(nu, sbw, si=11./3.):
    """
    Calculates the weak-strong transition frequency by scaling
    the scintillation bandwidth with nu and setting equal to nu.

    Input:
       nu = radio frequency of the input scintillation bandwidth (sbw)  (GHz)
       sbw = scintillation bandwidth (MHz)
    Output:
       transition frequency in GHz
    """
    if si == -2 or si == 2:
        print('transition_frequency_from_obs: si = %d'%(si), ' not allowed')
        return
    else:
        xsbw = (2 * si) / (si - 2)
        transition_frequency = nu * (1000. * nu / sbw)**(1./(xsbw-1))
        return transition_frequency

def taud_from_thetad(thetad, Deff):
    """
    Calculates pulse broading time from angular diameter.
    Works for any kind of medium (uniform, screen, Galactic scattering
    of extragalactic sources, etc.) by using an appropriate Deff. 

    Input:
        thetad = measured scattering diameter           mas 
        Deff = effective distance of scattering medium  kpc
               [e.g. for a Galactic layer of thickness Lg measured
                from the Sun,  Deff = Lg/2]
    Output:
        taud = pulse broadening time                    ms

    Method:
        Calculation relates the mean pulse broadening time to
        the mean-square scattering angle assuming a Gaussian
        scattered angular distribution. 

    Dependencies:
        Need c = speed of light and kpc defined in cgs units. 
        Need mas = milliarcsecond  defined.

    JMC 2020 Jan 1
    """
    taud = (Deff * kpc * (thetad * mas)**2) / (16 * np.log(2) * c) / ms      
    return taud

def rsigma(nu, SM, linner=100., si=11./3.):
    """
    NOT YET COMPLETED 

    Calculates the ratio R_\sigma(\beta) defined in pbcosmo.tex
    that is the ratio of the RMS 1D scattering angle to the RMS angle 
    given by the FWHM of the equivalent Gaussian fitted to the scattered
    image.     This ratio is \ge 1.

    Input:
       nu = RF                                  GHz
       SM = scattering measure                  kpc m^{-20/3}
       linner = inner scale                     km
       si = spectral index of wavenumber spectrum (11/3 for Kolmogorov)

    Output:
       Rsigma = ratio as defined above

    Requires:
       r_e = classical electron radius
       c = speed of light
    """

    beta4 = 4-si
    beta2 = beta-si

    coeff = (pi * (2*pi)**(beta4/2.)) / (np.sqrt(beta4) * 4.*pi**2 * fbeta(si))**(1./beta2)
    return


    


def sm_from_thetad_plane_wave_meansq_method(nu, thetad, si=11./3., linner=100.):
    """
    Calculates SM from angular broadening diameter
    for a power-law wavenumber spectrum with spectral index si == beta.
    [nb. beta not used here to avoid confusion with the beta PDF]

    This method uses the mean-square angle and generally depends
    (weakly) on the inner scale.

    From A-O book snippet (Drafted in 2019)
    'Estimating the Wavenumber Spectrum for Electron Density in the ISM'
    Eq 33.

    Note that the angular diameter as measured from the visibility function
    of a scattered source is substantially less than 
    [<\theta^2> / 2]^{1/2} for small linner and small scattering strength
    (with strength measured as \lambda^2 SM.   This is because of the wings
    on the image that extend to larger angles than for a Gaussian function.

    Thus proper use of this function requires relating thetad to
    the angular variance.    This relation depends on linner and
    \lambda^2 SM.  This is shown in the notes 'Cosmological Integrals
    for Dispersion and Scattering'  in pbcosmo.tex
    [Figure 4 shows the ratio of angles, R_{\sigma}].
  
    For linner = 100 km to 1000 km and SM/\nu^2 [\nu in GHz] \sim 10^{-4}, the
    RMS angle is larger than \thetad by a factor R_{\sigma} \sim 2.5 to 3.5.

    Input:
        nu = RF                                             GHz
        thetad = angular scattering diameter                mas
        si = 11./3. for Kolmogorov spectrum 
        linner = inner scale (default = 100 km)             km

    Output:
        scattering measure                                  kpc m^{-20/3}

    Requires definitions in cgs units of
        r_e = classical electron radius
        km = kilometer
        GHz = 10^9 Hz
        mas = milliarcsecond
        SMunit
    as in get_constants_waveprop_astropy_version

    JMC 2020 Jan 1
    """

    beta4 = si - 4
    wavelen = c / (nu * GHz) 

    print(si, beta4, wavelen, SMunit, mas)
    SM = (thetad * mas)**2 \
       / (4.*np.log(2.) * k_eta(si) * wavelen**4 * r_e**2 * (linner*km)**beta4)\
       / SMunit
    return SM

def sm_from_thetad_plane_wave_width_method(nu, thetad, si=11./3.):
    """
    Calculates SM from angular broadening diameter
    for a power-law wavenumber spectrum with spectral index si == beta.
    [nb. beta not used here to avoid confusion with the beta PDF]

    This method uses the width of the scattered image.
    In the strong but not super-strong scattering regime, 
    the SM estimate is independent of the inner scale. 

    From Cordes and Lazio (2003) NE2001 Paper II Eq. A10 with theta < theta_cross

    Result is consistent with Eq. A17. 

    Input:
        nu = RF                                             GHz
        thetad = angular scattering diameter                mas
        si = 11./3. for Kolmogorov spectrum 
    Output:
        scattering measure                                  kpc m^{-20/3}

    Requires definitions in cgs units of
        r_e = classical electron radius
        GHz = 10^9 Hz
        mas = milliarcsecond
        SMunit
    as in get_constants_waveprop_astropy_version

    JMC 2020 Jan 1
    """

    beta2 = si - 2.

    Cbeta = (pi / (2.*np.sqrt(np.log(2))))**beta2 \
          / (4*pi**2 * r_e**2 * c**si * fbeta(si))

    SM = Cbeta * (nu * GHz)**si * (thetad * mas)**beta2 \
       / SMunit
    return SM

def taux_from_thetax(d_mod,sm,smtau,smtheta,theta_x,tau,sbw):

	# Effective distance to dominant scattering  
    deffsm2 = d_mod*(sm - smtau/6. - smtheta/3.)/sm     

    # Alternative for extragalactic temporal broadening
    fwhm2sigma = 1. / sqrt(8.*np.log(2.))
    #taufactor = (((mas*fwhm2sigma)**2 * kpc) / (2.*c)) *  1000. # old - spurious factor of 2
    taufactor = (((mas*fwhm2sigma)**2 * kpc) / (c)) *  1000. # new - factor of 2 removed

    tau_x = taufactor * deffsm2 * theta_x**2            # ms
    sbw_x = sbw * tau / tau_x                           # MHz

    return tau_x,sbw_x,deffsm2


def sm_from_tau(nu, tau, Deff, si=11./3.):
    """
    Calculates SM from pulse broadening time and (effective) distance. 

    Input:
        nu = RF                                             GHz
        tau = pulse broadening time                         ms
        Deff = effective distance to scattering region      kpc
    Output:
        scattering measure                                  kpc m^{-20/3}
    """

    # NEED TO FINISH THIS

